#!/usr/bin/env python3
#-*- coding: utf-8 -*-

import ipaddress
import socket
import select
import sys
import threading

import raisin


class Client(threading.Thread):
    """
    Unique client TCP.
    Unique car il faut creer un objet par connection,
    chaque client est associe a un serveur.
    """
    def __init__(self, ip, port=0, signature=None):
        assert type(ip) is str, "L'ip doit etre une chaine de caractere."
        assert type(port) is int, "Le port doit etre un entier."
        assert port >= 0, "Le port doit etre positif ou nul."

        threading.Thread.__init__(self)
        self.signature = signature          # c'est la signature pour l'imprimente
        self.server_infos = {               # les informations concernant le serveur
            "ipv4_wan": None,
            "ipv4_lan": None,
            "ipv6": None,
            "dns_ipv4": None,
            "dns_ipv6": None,
            "port": None,
            "port_forwarding": None,
            "last_check": None,
        }
        self.server_in_lan = None           # booleen qui permet de dire si le serveur que l'on cherche a joindre est sur le meme reseau local que ce client

        with raisin.Printer("Initialisation of client..." , signature=signature) as p:
            # recherche des informations permetant de creer une connection
            
            # cas d'un nom de domaine raisin
            infos = raisin.communication.dns.get(ip, signature=self.signature)
            if infos: # si le serveur est deja recence par raisin
                self.server_infos = infos
                if infos["ipv4_lan"] != None and infos["ipv4_wan"] == str(raisin.Id().ipv4_wan): # si ce serveur a la meme ip publique
                    self.server_in_lan = True # c'est qu'il fait parti du meme reseau local
                    p.show("Server (lan) ipv4: %s, port: %d." % (infos["ipv4_lan"], infos["port"]))
                elif infos["port_forwarding"]: # si il est de toute facon accessible depuis l'exterieur
                    self.server_in_lan = False # on considere que le serveur est dehors
                    if infos["ipv6"]:
                        p.show("Server ipv6: %s, port: %d." % (infos["ipv6"], infos["port_forwarding"]))
                    if infos["ipv4_wan"]:
                        p.show("Server (wan) ipv4: %s, port: %d." % (infos["ipv4_wan"], infos["port_forwarding"]))
                else:
                    raise RuntimeError("Serveur non local, son port externe est inconu.")

            else:
                # cas d'un vrai nom de dommaine
                dns = ip
                ip_dns = raisin.communication.dns.is_domain(ip, signature=signature)
                if ip_dns:
                    ip = ip_dns
                    p.show("Domain %s translated in %s." % (dns, ip))

                # cas d'une ipv6 ou d'une ipv4
                if raisin.communication.dns.is_ipv6(ip):
                    self.server_infos["ipv6"] = ip
                    p.show("Server ipv6: %s." % ip)
                    if ip_dns:
                        self.server_infos["dns_ipv6"] = dns
                elif raisin.communication.dns.is_ipv4(ip):
                    p.show("Server ipv4: %s." % ip)
                    if ip_dns:
                        self.server_infos["dns_ipv4"] = dns
                else:
                    raise RuntimeError("Entree incorrecte, ce n'est ni une ip, ni un nom de domaine.")
                self.server_in_lan = ipaddress.ip_address(ip).is_private
                if not self.server_infos["ipv6"]:
                    if self.server_in_lan:
                        self.server_infos["ipv4_lan"] = ip
                        p.show("Server in (lan).")
                    else:
                        self.server_infos["ipv4_wan"] = ip
                        p.show("Server in (wan).")

                # le port
                if 0 == port:
                    raise RuntimeError("Le port doit etre precise.")
                if self.server_in_lan:
                    self.server_infos["port"] = port
                else:
                    self.server_infos["port_forwarding"] = port
                p.show("Port: %d." % port)

            # creation de la connection
            self.tcp_socket = None
            if self.server_infos["ipv6"]:
                try: # on tente d'abord en ipv6
                    self.tcp_socket = socket.socket(
                        socket.AF_INET6,        # socket internet en ipv6
                        socket.SOCK_STREAM)     # creation d'un TCP/IP socket, SOCK_STREAM => TCP
                    self.tcp_socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, False)
                    p.show("Socket initialised in ipv6.")
                except Exception as e: # si la connection en ipv6 a echouee
                    pass
            if self.tcp_socket == None and self.server_infos["ipv4_lan"] and self.server_infos["port"] or self.server_infos["ipv4_wan"] and self.server_infos["port_forwarding"]:
                try: # allors on essay en ipv4
                    self.tcp_socket = socket.socket(
                        socket.AF_INET,         # socket internet plutot qu'un socket unix
                        socket.SOCK_STREAM)     # creation d'un TCP/IP socket, SOCK_STREAM => TCP
                    self.tcp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1) # on tente de reutiliser le port si possible
                    p.show("Socket initialised in ipv4.")
                except Exception as e:
                    pass
            if self.tcp_socket == None:
                raise RuntimeError("Impossible de creer un socket TCP, peut etre n'y a-t-il pas internet?")

    def connect(self):
        """
        se connecte reelement au serveur
        """
        ip = None
        port = None
        with raisin.Printer("Connection au serveur...", signature=self.signature) as p:
            # recuperations des donnees
            if self.server_infos["ipv6"]:
                ip = self.server_infos["ipv6"]
                if self.server_in_lan:
                    port = self.server_infos["port"]
                else:
                    port = self.server_infos["port_forwarding"]
            elif self.server_infos["ipv4_wan"] or self.server_infos["ipv4_lan"]:
                if self.server_in_lan:
                    ip = self.server_infos["ipv4_lan"]
                    port = self.server_infos["port"]
                else:
                    ip = self.server_infos["ipv4_wan"]
                    port = self.server_infos["port_forwarding"]
            else:
                raise RuntimeError("Le serveur n'a pas d'ip.")
            if port == None:
                raise RuntimeError("Le port de connecvtion est inconu!")
            if ip == None:
                raise RuntimeError("L'ip n'est pas connue!")

            # connection veritable
            self.tcp_socket.connect((ip, port))


    def run(self):
        """
        Methode special appelle par la classe parente.
        Au moment ou l'on fait: self.start()
        """
        pass


"""
copier coller du site
"""

def example():
    # Choosing Nickname
    nickname = input("Choose your nickname: ")

    # Connecting To Server
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    client.connect(('127.0.0.1', 55555))

    # Listening to Server and Sending Nickname
    def receive():
        while True:
            try:
                # Receive Message From Server
                # If 'NICK' Send Nickname
                message = client.recv(1024).decode('ascii')
                if message == 'NICK':
                    client.send(nickname.encode('ascii'))
                else:
                    print(message)
            except:
                # Close Connection When Error
                print("An error occured!")
                client.close()
                break

    # Sending Messages To Server
    def write():
        while True:
            message = '{}: {}'.format(nickname, input(''))
            client.send(message.encode('ascii'))

    # Starting Threads For Listening And Writing
    receive_thread = threading.Thread(target=receive)
    receive_thread.start()

    write_thread = threading.Thread(target=write)
    write_thread.start()







#!/usr/bin/env python3
#-*- coding: utf-8 -*-

__all__ = ["answering", "dns", "dropbox", "mail", "tcp_client", "tcp_server"]

from . import answering
from . import dns
from . import dropbox
from . import mail
from . import tcp_client
from . import tcp_server



#!/usr/bin/env python3

"""
Serialise des objets pour les rendre transportables
"""

import cloudpickle
import inspect
import io
import itertools
import json
import os
import struct
import sys

from .constants import *
from .tools import *
from ..tools import Printer


def dumps(gen):
    """
    |===========================================|
    | Met sous forme de chaine de caractere les |
    | donnees serialisee par 'serialize'.       |
    |===========================================|

    Ne fait aucune verifications sur l'entree.
    Regroupe N_BYTES octets en N_SYMB caracteres ou
    bien 3 octets en 4 caracteres celon ce qui permet
    de compacter au mieux le resultat.

    parametres
    ----------
    :param gen: Generateur provenant directement de 'serialize'.
        Il ne doit pas etre vide ni epuise.
    :type gen: <class 'generator'>

    sortie
    ------
    :return: Une chaine de caractere contenant les objets serializes.
    :rtype: str
    """
    def anticipate_var(pack, size):
        """
        Generator qui cede des packets de 'size' octets.
        Cede aussi un booleen qui annonce l'arrive du dernier packet.
        """
        div, rest = divmod(len(pack), size)
        final = div + 1 if rest else div # ceil(pack/size)
        for i in range(final):
            yield (i+1 == final), pack[size*i:size*(i+1)]

    def int_to_alph(i, n_symb, n1):
        """
        Converti un entier en une chaine
        constitue par des symboles de l'alphabet.
        contraintes:
            i <= n1**n_symb
            retourne n_symb caracteres
        """
        chaine = ""
        while i:
            i, r = divmod(i, n1)
            chaine = ALPHABET[r] + chaine
        chaine = ALPHABET[0]*(n_symb-len(chaine)) + chaine
        return chaine

    def pack_to_int(pack):
        """
        Converti une chaine binaire en un entier.
        """
        return int.from_bytes(pack, byteorder="big", signed=False)

    # preparation (n1**n_symb >= n2**n_bytes)
    pack = b"".join(gen)
    if len(pack) < N_BYTES*3/4: # Si il vaut mieu ne pas faire des packets trop gros.
        sortie = HEADER["small dumps"][1].decode()
        n1 = 64 # Nbr de caracteres differents.
        n_bytes = 3 # C'est le nombre d'octet que l'on regroupe.
        n_symb = 4 # On converti 3 octets en 4 caracteres.
    else:
        sortie = HEADER["large dumps"][1].decode()
        n1 = len(ALPHABET) # Nbr de caracteres differents.
        n_bytes = N_BYTES # C'est le nombre d'octet que l'on regroupe.
        n_symb = N_SYMB # Regroupe N_BYTES octets en N_SYMB caracteres.

    # convertion
    for is_end, pack in anticipate_var(pack, n_bytes):
        sortie += int_to_alph(
            pack_to_int(pack),
            n_symb,
            n1)
        if is_end and len(pack) < n_bytes:
            return sortie + ALPHABET[len(pack)]
    return sortie

def serialize(obj, **kwargs):
    """
    |====================================================|
    | Serialise, compresse et encrypt tous type d'objet. |
    |====================================================|

    Ne fait aucune verifications des entrees.
    Tous les parametres doivent etre specifies.

    parametres
    ----------
    :param obj:
        N'importe quel type d'objet python, celui a serialiser.
    :type obj: object
    :param compresslevel:
        Permet de compresser sans perte le resultat.
    :type compresslevel: int
        -1: Intelligente, compresse au mieu
            en fonction du debit demande.
            Si 'parallelization_rate' >= 0
        0: Pas de compression du tout.
        1: Choisi un encodage moins lisible par
           l'humain mais plus performant.
        2: Legere compression, la plus rapide
           en temps de calcul.
        3: Legere compression qui maximise
           le ratio 'taille_gagnee/temps'.
        4: Moyenne compression qui vise un ratio
           environ egal a 95% du ratio maximum.
        5: Compression maximale qui ne se preocupe pas du temps.
    :param copy_file:
        Permet de copier le fichier si 'obj' est un nom de fichier.
        La chaine de caractere ne sera alteree en aucun cas.
    :type copy_file: bool
        True: Le fichier est copie et fait parti du resultat.
              Au moment de la deserialisation,
              le fichier sera regenere.
        False: Si 'obj' est un nom de fichier, il sera
               seulement considerer comme un 'str'.
    :param psw: Permet de chiffrer la valeur de retour.
    :type psw: str, bytes, None
        str: Chiffre avec l'algorithme AES.
        bytes: Clef publique au format PEM, RSA est utilise.
        None: Ne chiffre rien du tout.
    :param authenticity: Permet d'ajouter un hash cryptographique.
    :type authenticity: bool
        True: Ajoute des donnees de controle.
        False: N'ajoute rien du tout.
    :param parallelization_rate:
        Permet de partager le calcul afin de mieu profiter
        des resources de l'ordinateur.
    :type parallelization_rate: int
        0: Aucune parallelisation, tout est excecute
           dans le processus courant.
        1: Pseudo parallelisation, reparti les operations
           dans des threads du module 'threading'.
        2: Legere parallelisation, utilise les differents coeurs
           de la machine avec le module 'multiprocessing'.

    sortie
    ------
    :return: Une suite de 'bytes' qui contient l'information
             suffisante pour reconstiuer l'objet.
    :rtype: <class 'generator'>
    """
    with Printer("Serialization of a {} object...".format(type(obj).__name__)) as p:
        obj_ser = to_gen(gen=_serialize(
            obj,
            copy_file=kwargs["copy_file"],
            p=p,
            compact=(1 if kwargs["compresslevel"] else 0)))
        if kwargs["compresslevel"] not in (0, 1):
            import raisin.serialization.compress as compress
            obj_ser = to_gen(gen=compress.compress_gen(obj_ser, **kwargs))
        if kwargs["authenticity"]:
            import raisin.serialization.encrypt as encrypt
            obj_ser = to_gen(gen=encrypt.authenticity(
                obj_ser,
                compact=(1 if kwargs["compresslevel"] else 0)
                ))
        if kwargs["psw"]:
            import raisin.serialization.encrypt as encrypt
            obj_ser = to_gen(gen=encrypt.cipher_gen(
                obj_ser,
                psw=kwargs["psw"],
                compact=(1 if kwargs["compresslevel"] else 0),
                ))
        yield from obj_ser

def _serialize(obj, copy_file, p, compact):
    """
    |=========================================|
    | Serialise les objets sans les securiser |
    | ni sans non plus les compresser.        |
    |=========================================|
    
    parametres
    ----------
    :param obj: L'objet a serialiser.
    :type obj: object
    :param copy_file: Booleen qui indique si le nom de fichier doivent
        etre considerer comme des str ou comme un fichier.
        -True: Copie aussi le contenu du fichier.
        -False: Ne cherche rien a savoir, interprete come un str.
    :type copy_file: bool
    :param p: Instance de l'objet printer.
    :type p: raisin.tools.Printer
    :param compact: Determine si les objets serializes doivent restes humainement visibles.
    :type compact: int
        0: Le plus lisible possible.
        1: Le plus compact possible.
    sortie
    ------
    :return: L'objet serialise decoupe en packets de taille tres variable.
    :rtype: iter
    """
    def is_filename(obj, copy_file):
        """
        Retourne True si l'objet est un chemin vers un fichier
        et qu'il faut copier les fichiers.
        """
        if not copy_file:                   # si l'utilisateur ne nous autorise pas d'interpreter l'objet comme un fichier
            return False                    # alors on ne fait meme pas la suite du test
        if type(obj) == str:                # deja il faut que l'objet soit une chaine de caractere
            if len(obj) < 32767:            # et qu'il ne soit pas trop gros, sinon microchiotte windaube plante
                return os.path.isfile(obj)  # alors on fait le test
        return False    

    def is_jsonisable(obj, copy_file):
        """
        Retourne True si l'objet peu etre serialise avec json.
        Retourne False le cas echeant.
        """
        if type(obj) in (int, float, bool, type(None)):     # si l'objet est vraiment basique
            return True                                     # alors json sait le serialiser
        elif type(obj) == str:                              # dans le cas ou c'est une chaine de caractere
            if not is_filename(obj, copy_file):
                return True
        elif type(obj) == list:                             # dans le cas ou l'objet est une liste
            for element in obj:                             # on a pas d'autre choix que de parcourir la liste
                if not is_jsonisable(element, copy_file):   # si l'un des elements n'est pas compatible avec json
                    return False                            # alors la liste n'est pas jsonisable
            return True                                     # par contre, si tout est ok, allors on le jsonisifi
        elif type(obj) == dict:                             # de meme, dans le cas ou l'objet est un dictionnaire
            for key, value in obj.items():                  # pour faire la verification, on s'y prend comme pour les listes
                if not is_jsonisable(value, copy_file):     # si l'un des elements du dictionnaire fait tout capoter
                    return False                            # on ne va pas plus loin
                elif not is_jsonisable(key, copy_file):     # meme si c'est moin probable, mais si la clef n'est pas satisfesante
                    return False                            # tant pis, ça fait tout arreter quand meme
            return True                                     # si tout semble ok, et bien on dit qu'il est jsonalisable
        return False                                        # dans tous les autres cas, json ne fait pas bien le boulot

    def ser_filename(filename):
        """
        Encapsule le nom du fichier et le contenu du fichier.
        """
        filename_bytes = filename.encode("utf-8")
        yield size_to_tag(len(filename_bytes)) + filename_bytes
        with open(filename, "rb") as f:
            while 1:
                pack = f.read(BUFFER_SIZE)
                if pack:
                    yield pack
                else:
                    break

    def ser_float(f):
        """
        Recherche la representation la plus compacte possible.
        Retourne le type de solution ("s" ou "b") et les datas
        """
        as_str = str(f)                 # on represente le flotant en virant les 0 en trop
        if as_str[:2] == "0.":          # 0.123 devient .123
            as_str = as_str[1:]
        elif as_str[:3] == "-0.":       # -0.123 devient -.123
            as_str = as_str[0] + as_str[2:]
        elif as_str[-2:] == ".0":       # si il y a des 0 en trop a la fin
            as_str = as_str[:-2]        # 12.0 devient 12
            end_0 = 0 # nombre de 0 a la fin
            while as_str[-end_0-1:] == "0"*(end_0+1):
                end_0 += 1
            if end_0 >= 3:
                as_str = as_str[:-end_0] + "e" + str(end_0)
        as_str = as_str.replace("e+", "e")
        
        if len(as_str) < 16:
            if len(as_str) % 2:
                as_str = "0" + as_str
            pack = b""
            cor = {**{str(i):i for i in range(10)}, **{"-":10, "e":11, ".":12}}
            for i in range(0, len(as_str), 2):
                pack += bytes([cor[as_str[i+1]] + cor[as_str[i]]*len(cor)])
            return "round float", pack
        return "normal float", struct.pack("d", f)

    def ser_textiowrapper(file, p):
        """
        Cede des packet d'entete et de contenu du fichier
        """
        closed = b"\x01" if file.closed else b"\x00"     # 1 si le flux est ferme ou 0 si il est encore actif
        mode = file.mode.encode("utf-8")                 # son mode ("r", "w", "a", ...)
        encoding = file.encoding.encode("utf-8")         # l'encodage dans lequel on doit lire le fichier
        errors = file.errors.encode("utf-8")             # les differents modes de gestion possible des erreur de lecture
        stream_position = b"".join(_serialize(
            obj=(0 if obj.closed else obj.tell()),
            copy_file=False,
            p=p,
            compact=1))                                 # la position en nombre de caractere du curseur courant par rapport au debut du fichier
        
        yield size_to_tag(len(closed)) + closed \
              + size_to_tag(len(mode)) + mode \
              + size_to_tag(len(encoding)) + encoding \
              + size_to_tag(len(errors)) + errors \
              + size_to_tag(len(stream_position)) + stream_position

        if not b"r" in mode and not closed:             # si on est en ecriture
            file.flush()                                # on force l'ecriture sur le support phisique
        with open(file.name, "rb") as f:
            while 1:
                pack = f.read(BUFFER_SIZE)
                if pack:
                    yield pack
                else:
                    break

    def ser_buffered(file, p):
        """
        Cede les packets qui contiennent
        toutes l'informations suffisante pour
        recrer un fichier binaire dans le meme etat.
        """
        closed = b"\x01" if file.closed else b"\x00"    # 1 si le flux est ferme ou 0 si il est encore actif
        mode = file.mode.encode("utf-8")                # son mode ("rb", "wb", "ab", ...)
        stream_position = b"".join(_serialize(
            obj=(0 if file.closed else file.tell()),
            copy_file=False,
            p=p,
            compact=1))                                 # la position en nombre de caractere du curseur courant par rapport au debut du fichier
        
        yield size_to_tag(len(closed)) + closed \
              + size_to_tag(len(mode)) + mode \
              + size_to_tag(len(stream_position)) + stream_position

        if not b"r" in mode:                            # si on est en ecriture
            if not closed:
                file.flush()                            # on force l'ecriture sur le support phisique
            with open(file.name, "rb") as f:            # affin d'avoir acces au contenu
                while 1:
                    pack = f.read(BUFFER_SIZE)
                    if pack:
                        yield pack
                    else:
                        break
        else:                                           # si on est en lecture
            curs = file.tell()                          # on en profite pour ne pas reouvrir un autre fichier
            file.seek(0, 0)
            while 1:
                pack = file.read(BUFFER_SIZE)
                if pack:
                    yield pack
                else:
                    break
            file.seek(curs, 0)

    def ser_iter(iterable, copy_file, compact):
        """
        'iterable' est un iterable.
        Serialise un iterable on concatenant de facon
        reversible la serialisation de chacuns des elements.
        """
        with Printer("Iterable serialization...") as p:
            new = b"n"                   # fanion qui announce que il faut passer a l'element suivant
            end = b"e"                   # fanion qui indique que c'est fini
            cont = b"c"                  # fanion qui indique qu'on est encore en plein dans l'objet

            for is_end_iter, e in anticipate(iterable): # pour chaque element de l'iterable
                for is_end_elem, pack in anticipate(
                        to_gen(
                            gen=_serialize(
                                e,
                                copy_file=copy_file,
                                p=p,
                                compact=compact)
                            )
                        ):
                    indication = (end if is_end_iter else new) if is_end_elem else cont
                    yield size_to_tag(len(pack)+1) + pack + indication

    if type(obj) == int:
        if not compact and sys.getsizeof(obj) <= 64:
            p.show("small int -> data")
            yield HEADER["small int"][0] + str(obj).encode()
        else:
            p.show("large int -> data")
            yield HEADER["large int"][compact] + obj.to_bytes(
                length=(8 + (obj + (obj < 0)).bit_length()) // 8,
                byteorder="big",
                signed=True) # on encode plus efficacement
    elif is_filename(obj, copy_file):
        p.show("filename + file_content -> data")
        yield HEADER["filename"][compact]
        yield from ser_filename(obj)
    elif type(obj) == str:
        p.show("str -> data")
        yield HEADER["str"][compact] + obj.encode("utf-8")
    elif type(obj) == float:
        p.show("float -> data")
        if compact:
            g, data = ser_float(obj)
            yield HEADER[g][compact] + data
        else:
            yield HEADER["float"][compact] + str(obj).encode()
    elif type(obj) == complex:
        p.show("complex -> data")
        r = b"".join(_serialize(obj.real, copy_file=copy_file, p=p, compact=compact))
        i = b"".join(_serialize(obj.imag, copy_file=copy_file, p=p, compact=compact))
        yield HEADER["complex"][compact] + size_to_tag(len(r)) + r + i
    elif type(obj) == bytes:
        p.show("bytes -> data")
        yield HEADER["bytes"][compact] + obj
    elif issubclass(type(obj), io.TextIOWrapper):
        p.show("io.TextIOWrapper -> data")
        yield HEADER["TextIOWrapper"][compact]
        yield from ser_textiowrapper(obj, p)
    elif issubclass(type(obj), io.BufferedReader):
        p.show("io.BufferedReader -> data")
        yield HEADER["BufferedReader"][compact]
        yield from ser_buffered(obj, p)
    elif issubclass(type(obj), io.BufferedWriter):
        p.show("io.BufferedWriter -> data")
        yield HEADER["BufferedWriter"][compact]
        yield from ser_buffered(obj, p)
    elif type(obj) == tuple:
        p.show("tuple -> data")
        yield HEADER["tuple"][compact]
        yield from ser_iter(obj, copy_file, compact)
    elif type(obj) == bool:
        p.show("bool -> data")
        yield HEADER["true"][compact] if obj else HEADER["false"][compact]
    elif obj == None:
        p.show("null -> data")
        yield HEADER["null"][compact]
    elif type(obj) == set:
        p.show("set -> data")
        yield HEADER["set"][compact]
        yield from ser_iter(obj, copy_file, compact)
    elif type(obj) == frozenset:
        p.show("frozenset -> data")
        yield HEADER["frozenset"][compact]
        yield from ser_iter(obj, copy_file, compact)
    elif inspect.isclass(obj):                                      # si l'objet est une classe, qu'elle soit integree ou creee dans du code python
        p.show("class -> data")
        raise NotImplementedError()
    elif inspect.ismodule(obj):                                     # si l'objet est un module
        p.show("module -> data")
        raise NotImplementedError()
    elif inspect.isfunction(obj):                                   # si l'objet est une fonction python, qui inclut des fonctions creees par une expression lambda
        p.show("function -> data")
        raise NotImplementedError()
    elif inspect.isgenerator(obj):                                  # si l'objet est un generateur
        p.show("generator -> data")
        yield HEADER["generator"][compact]
        yield from ser_iter(obj, copy_file, compact)
    elif not compact and sys.getsizeof(obj) <= BUFFER_SIZE and is_jsonisable(obj, copy_file):
        p.show("json -> data")
        yield HEADER["json"][compact] + json.dumps(obj).encode("utf-8")
    
    elif type(obj) == list:
        p.show("list -> data")
        yield HEADER["list"][compact]
        yield from ser_iter(obj, copy_file, compact)
    elif type(obj) == dict:
        p.show("dict -> data")
        yield HEADER["dict"][compact]
        yield from ser_iter(
            obj.items(),
            copy_file,
            compact)
    else:
        p.show("unknown object -> pickle -> data")
        yield HEADER["pickle"][compact] + cloudpickle.dumps(obj)



#!/usr/bin/env python3

"""
|========================================================|
| Gestion de la partie application logicielle de raisin. |
|========================================================|

1) Le partage des ressources
-C'est dans ce sous-module que l'on gere
les ressources que l'on offre a la communautee.
2) La configuration
-C'est aussi ici que l'on interagit graphiquement
avec l'utilisateur afin de tout configurer.
-On l'aide a naviger dans les repertoires.
3) persistance des donnes
-Enfin, c'est ici que l'on s'occupe d'enregistrer les
parametres et certaines donnees.
"""

__all__ = ["hmi", "settings", "install", "uninstall"]



#!/usr/bin/env python3

"""
Gestion de la persistance des parametres de configuration
de l'application.
C'est ici que l'on enregistre les parametres et que
l'on va les extraire aussi.
"""

import copy
import os
import pprint
import re
import socket
import sys
import time
try:
    import psutil
except ImportError:
    psutil = None

from ..tools import Printer, id_, get_temperature
from .hmi import dialog
from .hmi import checks
from ..communication import constants

class Settings:
    """
    Permet d'acceder et de modifier les parametres.
    """
    def __init__(self, *, home=os.path.expanduser("~"), action="normal"):
        assert action in ("paranoiac", "normal", "altruistic", "custom"), \
            "Les actions ne peuvent que etre 'paranoiac', " \
            + "'normal', 'altruistic' ou 'custom'. Pas '%s'." % action
        
        self.home = home
        self.action = action # facon dont on doit choisir les parametre par defaut
        self.last_read = 0 # date de la derniere lecture sur le support physique.
        self.settings = {} # les parametres dictionaire
        self.original_settings = {} # les parametres initiaux

    def _load_settings(self):
        """
        Recupere le contenu du fichier 'settings.py'
        Retourne ce contenu deserialise
        ou bien renvoi {} si le fichier n'est pas existant ou corrompu.
        """
        if time.time() - self.last_read > 60*10:
            self.last_read = time.time()
            return load_settings(home=self.home)
        return self.settings
    
    def _dump_settings(self):
        """
        Enregistre les paramettres "self.settings".
        Attention, si self.settings n'est pas initalise,
        Tous les parametres sont perdus.
        """
        settings = copy.deepcopy(self.settings)
        if hasattr(self, "dns_ipv6") and self.dns_ipv6:
            settings["server"]["dns_ipv6"] = self.dns_ipv6
        if hasattr(self, "dns_ipv6") and self.dns_ipv4:
            settings["server"]["dns_ipv4"] = self.dns_ipv4
        dump_settings(settings, home=self.home)

    def _is_complete(self, settings, ref=None):
        """
        Retourne True si tous les champs sont presents.
        Retourne False si il y a un champ de trop ou qu'il en manque un.
        """
        model = {
         'account': {'automatic_update': None,
                     'email': None,
                     'give_activity_schedules': None,
                     'give_cpu_usage': None,
                     'give_internet_activity': None,
                     'give_ram_usage': None,
                     'padlock': {'break_time': None,
                                 'cipher': None,
                                 'notify_by_email': None,
                                 'paths': {'excluded_paths': None,
                                           'paths': None}},
                     'security': {'hash': None,
                                  'private_key': None,
                                  'psw': None,
                                  'sel': None,
                                  'public_key': None,
                                  'sentence_memory': None},
                     'username': None},
         'cluster_work': {'downflow': None,
                          'free_size': None,
                          'limit_bandwidth': None,
                          'limit_cpu_usage': None,
                          'limit_fan_noise': None,
                          'limit_ram_usage': None,
                          'low_cpu_usage': None,
                          'maximum_temperature': None,
                          'recording_directory': None,
                          'restrict_access': None,
                          'rising_flow': None,
                          'schedules_bandwidth': {'friday': None,
                                                  'monday': None,
                                                  'saturday': None,
                                                  'sunday': None,
                                                  'thursday': None,
                                                  'tuesday': None,
                                                  'wednesday': None},
                          'schedules_cpu_usage': {'friday': None,
                                                  'monday': None,
                                                  'saturday': None,
                                                  'sunday': None,
                                                  'thursday': None,
                                                  'tuesday': None,
                                                  'wednesday': None},
                          'schedules_fan_noise': {'friday': None,
                                                  'monday': None,
                                                  'saturday': None,
                                                  'sunday': None,
                                                  'thursday': None,
                                                  'tuesday': None,
                                                  'wednesday': None},
                          'schedules_ram_usage': {'friday': None,
                                                  'monday': None,
                                                  'saturday': None,
                                                  'sunday': None,
                                                  'thursday': None,
                                                  'tuesday': None,
                                                  'wednesday': None}},
         'server': {'accept_new_client': None,
                    'access_token': None,
                    'dns_ipv4': None,
                    'dns_ipv6': None,
                    'force_authentication': None,
                    'listen': None,
                    'network_name': None,
                    'port': None,
                    'port_forwarding': None}}
        if ref == None:
            ref = model

        for clef in ref:
            if clef not in settings:
                return False
            if ref[clef] != None:
                if not self._is_complete(settings[clef], ref=ref[clef]):
                    return False
        for clef in settings:
            if clef not in ref:
                return False
        return True

    def get_settings(self):
        """
        Si des parametres sont deja enregistres, il sont juste completes.
        Sinon, les parametres par defaut sont renvoyes.
        """
        settings = self._load_settings()                            # on recupere les parametres existants
        if not self._is_complete(settings):
            with Printer("Account settings...") as p:
                settings["account"] = settings.get("account", {})

                p.show("Load Username")
                settings["account"]["username"] = str(settings["account"].get("username", id_().username))
                if self.action == "custom":
                    settings["account"]["username"] = dialog.question_reponse(
                        "Username :",
                        default=settings["account"]["username"],
                        validatecommand=checks.username_verification)
                
                p.show("Load Email")
                if "email" not in settings["account"]:
                    import raisin
                    settings["account"]["email"] = str(settings["account"].get("email",
                        re.search(r"[a-zA-Z0-9._-]+@[a-zA-Z0-9._-]{2,}\.[a-z]{2,4}", raisin.__author__).group()))
                if self.action == "custom":
                    settings["account"]["email"] = dialog.question_reponse(
                        "Email :",
                        default=settings["account"]["email"],
                        validatecommand=checks.email_verification)
                
                p.show("Load Security")
                if "security" not in settings["account"]:             # si il n'y a pas de clef publique ou de clef privee
                    import raisin.security as security
                    private_key, public_key = security.rsa_keys()     # on commence par generer le couple de clef
                    settings["account"]["security"] = {
                        "private_key": private_key,
                        "public_key": public_key,
                        "hash": None,                                     # on ne va pas choisir un mot de passe
                        "psw": None,                                      # a la place de l'utilisateur
                        "sel": b"",                                        # donc on n'en met pas
                        "sentence_memory": ""}
                if self.action in ("custom", "paranoiac"):                 # si il faut tout de metre en metre un
                    raise NotImplementedError("Il faut le coder.")

                p.show("Load Vie privee")
                if self.action == "custom":
                    settings["account"]["give_internet_activity"] = dialog.question_binaire(
                        "Donner des informations concerant ma connection a internet ?")
                elif self.action == "paranoiac" or self.action == "normal":
                    settings["account"]["give_internet_activity"] = False
                elif self.action == "altruistic":
                    settings["account"]["give_internet_activity"] = True
                else:
                    settings["account"]["give_internet_activity"] = settings["account"].get("give_internet_activity", False)
                if self.action == "custom":
                    settings["account"]["give_activity_schedules"] = dialog.question_binaire(
                        "Donner des informations concerant l'alimentation de mon ordinateur ?")
                elif self.action == "paranoiac" or self.action == "normal":
                    settings["account"]["give_activity_schedules"] = False
                elif self.action == "altruistic":
                    settings["account"]["give_activity_schedules"] = True
                else:
                    settings["account"]["give_activity_schedules"] = settings["account"].get("give_activity_schedules", False)
                if self.action == "custom":
                    settings["account"]["give_cpu_usage"] = dialog.question_binaire(
                        "Donner des informations concerant la sollicitation du CPU ?")
                elif self.action == "paranoiac":
                    settings["account"]["give_cpu_usage"] = False
                elif self.action == "normal" or self.action == "altruistic":
                    settings["account"]["give_cpu_usage"] = True
                else:
                    settings["account"]["give_cpu_usage"] = settings["account"].get("give_cpu_usage", True)
                if self.action == "custom":
                    settings["account"]["give_ram_usage"] = question_binaire(
                        "Donner des informations concerant l'utilisation de la RAM ?")
                elif self.action == "paranoiac":
                    settings["account"]["give_ram_usage"] = False
                elif self.action == "normal" or self.action == "altruistic":
                    settings["account"]["give_ram_usage"] = True
                else:
                    settings["account"]["give_ram_usage"] = settings["account"].get("give_ram_usage", True)
                if self.action == "custom":
                    settings["account"]["automatic_update"] = question_binaire(
                        "Faire les mises a jours automatiquement ?")
                elif self.action == "paranoiac":
                    settings["account"]["automatic_update"] = False
                elif self.action == "normal" or self.action == "altruistic":
                    settings["account"]["automatic_update"] = True
                else:
                    settings["account"]["automatic_update"] = settings["account"].get("automatic_update", True)

                p.show("Load Padlock")
                if not "padlock" in settings["account"]:               # si il n'y a rien pour la gestion de l'antivol
                    settings["account"]["padlock"] = {
                        "cipher":False,                                     # chiffre les donnees personnelles et demande le mot de passe pour les remetres en clair
                        "paths":                                            # ce dictionaire contient les repertoire achiffrer et leur options
                            {
                            "paths": [],
                            "excluded_paths": [],
                            },
                        "break_time":3600,                                  # temps de pause ou l'on ne regarde pas l'ip
                        "notify_by_email":False,                            # m'envoyer un mail contenant l'ip, l'heure (le son et une image/video) l'orsque l'on chiffre les donnees perso
                    }

            with Printer("Cluster work settings...") as p:
                settings["cluster_work"] = settings.get("cluster_work", {})

                p.show("Load fan nose")
                if get_temperature() is None:    # si il n'est pas possible de recuperer la temperature
                    settings["cluster_work"]["limit_fan_noise"] = False
                else:
                    if self.action == "altruistic":
                        settings["cluster_work"]["limit_fan_noise"] = False
                    elif self.action == "custom":
                        settings["cluster_work"]["limit_fan_noise"] = checks.question_binaire(
                            "Voulez-vous limiter le bruit du ventilateur?", default=True)
                    else:
                        settings["cluster_work"]["limit_fan_noise"] = settings["cluster_work"].get("limit_fan_noise", True)
                settings["cluster_work"]["schedules_fan_noise"] = settings["cluster_work"].get("schedules_fan_noise",
                    {
                    "monday" : {"8:00":False, "22:00":True},
                    "tuesday" : {"8:00":False, "22:00":True},
                    "wednesday" : {"8:00":False, "22:00":True},
                    "thursday" : {"8:00":False, "22:00":True},
                    "friday" : {"8:00":False, "22:00":True},
                    "saturday" : {"10:00":False, "23:00":True},
                    "sunday" : {"10:00":False, "23:00":True},
                    })
                settings["cluster_work"]["maximum_temperature"] = settings["cluster_work"].get("maximum_temperature", 60) # temperature qui rend l'ordinateur bruyant

                p.show("Load CPU restriction")
                if self.action == "altruistic":
                    settings["cluster_work"]["limit_cpu_usage"] = False
                elif self.action == "custom":
                    settings["cluster_work"]["limit_cpu_usage"] = question_binaire(
                        "Voulez-vous limiter le taux de CPU?", default=True)
                else:
                    settings["cluster_work"]["limit_cpu_usage"] = settings["cluster_work"].get("limit_cpu_usage", True)
                settings["cluster_work"]["low_cpu_usage"] = settings["cluster_work"].get("low_cpu_usage", False) # on ne peu pas juste le metre a True, car il a beaucoup de choses a faire
                taux = 50 if os.cpu_count() <= 1 else int(100*(os.cpu_count() - 1)/os.cpu_count()) # pour la regulation du CPU, on laisse 1 coeur libre
                settings["cluster_work"]["schedules_cpu_usage"] = settings["cluster_work"].get("schedules_cpu_usage",
                    {
                    "monday" : {"8:00":taux, "22:00":100},
                    "tuesday" : {"8:00":taux, "22:00":100},
                    "wednesday" : {"8:00":taux, "22:00":100},
                    "thursday" : {"8:00":taux, "22:00":100},
                    "friday" : {"8:00":taux, "22:00":100},
                    "saturday" : {"10:00":taux, "23:00":100},
                    "sunday" : {"10:00":taux, "23:00":100},
                    })

                p.show("Load RAM restriction")
                if self.action == "altruistic":
                    settings["cluster_work"]["limit_ram_usage"] = False
                elif self.action == "custom":
                    settings["cluster_work"]["limit_ram_usage"] = checks.question_binaire(
                        "Voulez-vous limiter l'acces de la RAM?", default=True)
                else:
                    settings["cluster_work"]["limit_ram_usage"] = settings["cluster_work"].get("limit_ram_usage", True)
                maxi_virtual = 2*2**30 if not psutil else psutil.virtual_memory().total/2**20
                maxi_swap = 4*2**30 if not psutil else (psutil.swap_memory().total + psutil.virtual_memory().total)/2**20
                settings["cluster_work"]["schedules_ram_usage"] = settings["cluster_work"].get("schedules_ram_usage",
                    {
                    "monday" : {"8:00":int(maxi_virtual*0.75), "22:00":int(maxi_swap*0.9)},
                    "tuesday" : {"8:00":int(maxi_virtual*0.75), "22:00":int(maxi_swap*0.9)},
                    "wednesday" : {"8:00":int(maxi_virtual*0.75), "22:00":int(maxi_swap*0.9)},
                    "thursday" : {"8:00":int(maxi_virtual*0.75), "22:00":int(maxi_swap*0.9)},
                    "friday" : {"8:00":int(maxi_virtual*0.75), "22:00":int(maxi_swap*0.9)},
                    "saturday" : {"10:00":int(maxi_virtual*0.75), "23:00":int(maxi_swap*0.9)},
                    "sunday" : {"10:00":int(maxi_virtual*0.75), "23:00":int(maxi_swap*0.9)},
                    })

                p.show("Load bandwidth restriction")
                if self.action == "paranoiac":
                    settings["cluster_work"]["limit_bandwidth"] = True
                elif self.action == "custom":
                    settings["cluster_work"]["limit_bandwidth"] = checks.question_binaire(
                        "Voulez-vous limiter la bande passante?", default=False)
                else:
                    settings["cluster_work"]["limit_bandwidth"] = settings["cluster_work"].get("limit_bandwidth", False)
                settings["cluster_work"]["schedules_bandwidth"] = settings["cluster_work"].get("schedules_bandwidth",
                    settings["cluster_work"]["schedules_fan_noise"])               # on met par defaut les memes horaires de limiation que celle du ventillateur
                settings["cluster_work"]["downflow"] = settings["cluster_work"].get("downflow", 0.5) # debit descendant en Mio/s
                settings["cluster_work"]["rising_flow"] = settings["cluster_work"].get("rising_flow", 0.1) # debit montant en Mio/s

                p.show("Load recording directory")
                settings["cluster_work"]["recording_directory"] = settings["cluster_work"].get("recording_directory",
                    os.path.join(os.path.expanduser("~"), ".raisin"))   # c'est le repertoire dans lequel on enregistre les resultats
                if self.action == "paranoiac":
                    settings["cluster_work"]["free_size"] = 50_000
                elif self.action == "altruistic":
                    settings["cluster_work"]["free_size"] = 1000
                else:
                    settings["cluster_work"]["free_size"] = settings["cluster_work"].get("free_size", 10_000) # place a laisser disponible en Mio

                p.show("Load restrict acces")
                if self.action == "paranoiac":
                    settings["cluster_work"]["restrict_access"] = True
                elif self.action == "custom":
                    settings["cluster_work"]["restrict_access"] = checks.question_binaire(
                        "Voulez-vous restreindre les droits de l'application?", default=False)
                else:
                    settings["cluster_work"]["restrict_access"] = settings["cluster_work"].get("restrict_access", False)

            with Printer("Server settings...") as p:
                settings["server"] = settings.get("server", {})

                p.show("Load port information")
                if "port" not in settings["server"]:
                    with Printer("Search for an open local port..."):
                        for port in range(16384, 49152, 1): # range(1024, 49152, 1) mais c'est pour free
                            if port not in constants.RESERVED_PORTS and checks.port_verification(port):
                                settings["server"]["port"] = settings["server"]["port"] = port
                                break
                
                p.show("Load listen")
                settings["server"]["listen"] = settings["server"].get("listen", 2*os.cpu_count()) # par defaut, le serveur prend en charge 2 connection par thread

                p.show("Load network name")
                settings["server"]["network_name"] = settings["server"].get("network_name", "main")

                p.show("Load DNS")
                settings["server"]["dns_ipv6"] = settings["server"].get("dns_ipv6", "")
                settings["server"]["dns_ipv4"] = settings["server"].get("dns_ipv4", "")
                self.dns_ipv6 = settings["server"]["dns_ipv6"] # sauvegarde pour ne pas
                self.dns_ipv4 = settings["server"]["dns_ipv4"] # les perdres a l'enregistrement
                if settings["server"]["dns_ipv6"]:
                    if not checks.dns_ip_verification(settings["server"]["dns_ipv6"], 6):
                        settings["server"]["dns_ipv6"] = None
                if settings["server"]["dns_ipv4"]:
                    if not checks.dns_ip_verification(settings["server"]["dns_ipv4"], 6):
                        settings["server"]["dns_ipv4"] = None
                if not settings["server"]["dns_ipv6"] or not settings["server"]["dns_ipv4"]:
                    myid = id_()
                    if myid.ipv6 != None:
                        dns_ipv6 = socket.getfqdn(str(myid.ipv6))
                        if checks.dns_ip_verification(dns_ipv6, 6) and not settings["server"]["dns_ipv6"]:
                            settings["server"]["dns_ipv6"] = dns_ipv6
                    if myid.ipv4_wan != None:
                        dns_ipv4 = socket.getfqdn(str(myid.ipv4_wan))
                        if checks.dns_ip_verification(dns_ipv4, 4) and not settings["server"]["dns_ipv4"]:
                            settings["server"]["dns_ipv4"] = dns_ipv4

                p.show("Load port forwarding")
                settings["server"]["port_forwarding"] = settings["server"].get("port_forwarding", None)

                p.show("Load preferences")
                if self.action == "paranoiac":
                    settings["server"]["accept_new_client"] = False
                    settings["server"]["force_authentication"] = True
                elif self.action == "custom":
                    settings["server"]["accept_new_client"] = dialog.question_binaire(
                        "Demander mon autorisation avant d'accepter de nouveaux clients?", default=True)
                    settings["server"]["force_authentication"] = dialog.question_binaire(
                        "Forcer les clients a s'authentifier?", default=False)
                else:
                    settings["server"]["accept_new_client"] = settings["server"].get("accept_new_client", True)
                    settings["server"]["force_authentication"] = settings["server"].get("force_authentication", False)

                p.show("Load access token")
                settings["server"]["access_token"] = settings["server"].get("access_token", None)

        return settings

    def get_original(self, *clefs):
        """
        Recupere le parametre initial.
        """
        self.__call__()
        element = self.original_settings
        for clef in clefs:
            element = element[clef]
        return element

    def __call__(self):
        """
        Charge les paremetres pour les
        stocker dans l'attribut self.settings.
        """
        self.settings = self.get_settings()
        if not self.original_settings:
            self.original_settings = copy.deepcopy(self.settings)

    def __str__(self):
        self.__call__()
        return str(self.settings)

    def __repr__(self):
        self.__call__()
        return repr(self.settings)

    def __getitem__(self, clef):
        """
        Permet d'aller recuperer un parametre.
        """
        self.__call__()
        return self.settings[clef]

    def __setitem__(self, clef, item):
        """
        Permet d'aller modifier un parametre.
        Ne fait aucune verification.
        """
        self.__call__()
        self.settings[clef] = item
        self._dump_settings()
            
    def flush(self):
        """
        Force a ecrir l'etat actuel dans le disque.
        """
        if not self.settings:
            self.settings = self.get_settings()
        self._dump_settings()

def load_settings(*, home=os.path.expanduser("~")):
    """
    recupere le contenu du fichier 'settings.py'
    retourne ce contenu deserialise
    ou bien renvoi {} si le fichier n'est pas existant ou corrompu
    'home' est le repertoire courant de l'utilisateur.
    """
    assert isinstance(home, str), \
        "'home' have to be 'str', not %s." % type(home).__name__
    assert os.path.isdir(home), \
        "'home' have to be a repository. " \
        + "%s is not an existing repository." % repr(home)

    with Printer(
            "Load settings for {}...".format(
            repr(os.path.basename(home)))
        ) as p:
        rep = os.path.join(home, ".raisin")
        filename = "settings.py"
        if os.path.isdir(rep):
            if os.path.isfile(os.path.join(rep, filename)):
                try:
                    with open(os.path.join(rep, filename), "r", encoding="utf-8") as f:
                        return eval(f.read())
                except KeyboardInterrupt as e:
                    raise e from e
                except:
                    p.show("No valid file found")
                    return {}
        p.show("No valid file found")
        return {}

def dump_settings(settings, *, home=os.path.expanduser("~")):
    """
    enregistre les parametres
    ne fait aucune verification sur ces parametres
    'home' est le repertoire courant de l'utilisateur.
    """
    assert isinstance(home, str), \
        "'home' have to be 'str', not %s." % type(home).__name__
    assert os.path.isdir(home), \
        "'home' have to be a repository. " \
        + "%s is not an existing repository." % repr(home)

    with Printer("Save settings..."):
        rep = os.path.join(home, ".raisin")
        if not os.path.isdir(rep):              # si le repertoire raisin n'existe pas
            _make_tree(rep) # il est cree
        if "cluster_work" in settings:
            if "recording_directory" in settings["cluster_work"]:
                if not os.path.isdir(os.path.join(settings["cluster_work"]["recording_directory"], "results")):
                    _make_tree(
                        os.path.join(settings["cluster_work"]["recording_directory"], "results")
                        )
        
        with open(os.path.join(rep, "settings.py"), "w", encoding="utf-8") as f:
            stdcourant = sys.stdout
            sys.stdout = f
            try:
                pprint.pprint(settings)
            except Exception as e:
                raise e from e
            finally:
                sys.stdout = stdcourant

def _make_tree(tree_path, indentation=0):
    """
    Cree recursivement le chemin complet 'tree_path'
    si plusieurs repertoires doivent etre crees,
    Ils sont fait recursivement.
    Si le repertoire pointe vers un fichier, le fichier
    est degome pour laisser la place a un repertoire.
    """
    if os.path.isdir(tree_path):
        return
    if os.path.isfile(tree_path):
        with Printer("Deleted '%s'..." % tree_path):
            os.remove(tree_path)
    dirname = os.path.dirname(tree_path)
    with Printer("Make '%s'..." % tree_path):
        if not os.path.isdir(dirname):
            _make_tree(dirname, indentation=indentation+1)
        os.mkdir(tree_path)

settings = Settings()



#!/usr/bin/env python3

"""
|=========================|
| Human Machine Interface |
|=========================|

Interaction avec l'utilisation.
1) Permet de lui poser question soit directement
dans le terminal ou bien si tkinter est installe,
Un interface graphique plus poussee est employée.
2) Permet aussi de lui presenter les choses de facon 
humainement lisibles. Afin de lui presenter
les donnees de facon lisible.
"""

__all__ = ["dialog", "theme", "checks", "pathmanager", "schedules"]

import logging
try:
    import tkinter
    import tkinter.messagebox as messagebox
    import tkinter.ttk as ttk
except ImportError:
    logging.warn("'tkinter' failed to import, no possibility to have interface graphical.")
    tkinter = None
    messagebox = None
    ttk = None



#!/usr/bin/env python3

"""
Dechiffre les donnees avec RSA ou AES.
"""

from Cryptodome.Cipher import AES
from Cryptodome.Cipher import PKCS1_OAEP
from Cryptodome.PublicKey import RSA
import hashlib
import itertools
import os
import re

import raisin
import raisin.serialization.deserialize as deserialize
from .constants import *
from .tools import *
from ..tools import Printer
import raisin.security as security


class DecryptError(Exception):
    """
    Erreur leve l'orsqu'un mot de passe ou une clef n'est pas correcte.
    Ou lorsque l'authenticite n'est pas verifiee.
    """
    pass


def decipher(data, *, psw=None):
    r"""
    |================================================|
    | Dechiffre les donnees prealablement chiffrees. |
    |================================================|

    Ne fait aucune verification.
    L'entete doit etre comprise dans 'data'.

    parametres
    ----------
    :param data: Donnees chiffrees.
    :type data: bytes
    :param psw: Le mot de passe ou la clef publique rsa.
    :type psw: str, bytes, None
        str: Dechiffrage AES.
        bytes: Clef privee RSA.
        None: Dechiffrage qui utilise la clef privee de raisin.

    sortie
    ------
    :return: Les donnees dechiffrees.
    :rtype: bytes

    exemple
    -------
    :Example:
    >>> from raisin.serialization.cipher import cipher
    >>> from raisin.serialization.decipher import decipher
    >>>
    >>> mess_ciph = cipher(b"message en clair")
    >>> mess_ciph
    b'\x03\x08\x82\x08\x80:\x8c\xbe=\x07\xe92gy\x8a\x8
        ...
    \xbcr\xe4\xf4\x1a\xa1xr\xedGcF\x0c\xc2\xf4$\xad9H,\xe6\x8e'
    >>> decipher(mess_ciph)
    b'message en clair'
    >>>
    """
    if psw == None:
        import raisin.security
        psw = raisin.security.get_private_key()

    return b"".join(decipher_gen((
        lambda d: (yield d))(data),
        psw=psw
        ))

def decipher_aes(data, passphrase):
    """
    |========================================================|
    | Dechiffre des donnees chiffrees avec l'algorithme AES. |
    |========================================================|

    Ne fait aucune verification sur les entrees.
    L'entete doit faire partie de data.

    parametres
    ----------
    :param data: Donnees chiffrees.
    :type data: bytes
    :param passphrase: Extrait du mot de passe de 32 octets.
    :type passphrase: bytes

    sortie
    ------
    :return: les donnees dechiffrees 
    :rtype: bytes

    exemple
    -------
    :Example:
    >>> import hashlib
    >>> from raisin.serialization.encrypt import cipher_aes
    >>> from raisin.serialization.decrypt import decipher_aes
    >>>
    >>> passphrase = hashlib.blake2s(b'mot de passe', digest_size=32).digest()
    >>> ciph_mes = cipher_aes(b"message en clair", passphrase, 1)
    >>> decipher_aes(ciph_mes, passphrase)
    b'message en clair'
    >>>
    """
    with Printer("Deciphering bloc with AES..."):
        head, gen, data = get_header(pack=data)
        if head != "aes bloc":
            raise ValueError(
                "L'entete n'est pas bonne 'aes bloc' est attendu " \
                + "mais %s est recut." % head)
        data = data + b"".join(gen)
        nonce, tag, ciphertext = data[:16], data[16:32], data[32:]
        cipher = AES.new(passphrase, AES.MODE_EAX, nonce)
        try:
            return cipher.decrypt_and_verify(ciphertext, tag)
        except ValueError as e:
            raise DecryptError("La passphrase n'est pas la bonne.") from e

def decipher_rsa(data, private_key_pem):
    r"""
    |========================================================|
    | Dechiffre des donnees chiffrees avec l'algorithme RSA. |
    |========================================================|

    Ne fait pas de verification sur les entrees.

    parametres
    ----------
    :param data: Donnees chiffrees
    :type data: bytes
    :param private_key_pem: Clef privee serialisee au format PEM.
    :type private_key_pem: bytes

    sortie
    ------
    :return: les donnees dechiffrees 
    :rtype: bytes

    exemple
    -------
    :Example:
    >>> from raisin.security import rsa_keys
    >>> from raisin.serialization.encrypt import cipher_rsa
    >>> from raisin.serialization.decrypt import decipher_rsa
    >>> priv, pub = rsa_keys()
    >>> ciph_mes = cipher_rsa(b"message en clair", pub, 1)
    >>> ciph_mes
    b'3\x08\x80&\x1d4g\xb7_\xc7\xfd\x9d% ... \xc0\x89\xbaI\xeb\xa28\xa4'
    >>> decipher_rsa(ciph_mes, priv)
    b"message en clair"
    >>>
    """
    with Printer("Deciphering block with RSA..."):
        head, gen, pack = get_header(pack=data)
        if head != "rsa bloc":
            raise ValueError(
                "L'entete n'est pas bonne 'rsa bloc' est attendu " \
                + "mais %s est recut." % head)
        
        try:
            decryptor = PKCS1_OAEP.new(RSA.importKey(private_key_pem))
        except ValueError as e:
            raise DecryptError("La clee privee est chiffree!") from e
        try:
            return b"".join(
                    map(
                        decryptor.decrypt,
                        deconcat_gen(
                            gen=gen,
                            pack=pack
                        )
                    )
                )
        except ValueError as e:
            raise DecryptError("La clef privee n'est pas la bonne.") from e

def decipher_gen(gen, psw):
    """
    |===========================================|
    | Dechiffre sequentiellement chaque packets |
    | cedes par le generateur en les regroupant |
    | tel qu'ils etaient avant.                 |
    |===========================================|

    parametres
    ----------
    :param gen: Generateur qui cede des 'bytes'.
    :type gen: iter
    :param psw: Le mot de passe ou la clef publique rsa.
    :type psw: str, bytes, None
        str: Dechiffrage rapide avec AES.
        bytes: Clef privee RSA.

    sortie
    ------
    :return: Les donnees dechiffrees.
    :rtype: iter

    exemple
    -------
    :Example:
    >>> import uuid
    >>> from raisin.serialization.encrypt import cipher_gen
    >>> from raisin.serialization.decrypt import decipher_gen
    >>>
    >>> gen1 = [uuid.uuid4().bytes for i in range(10)]
    >>> gen2 = cipher_generator(gen1)
    >>> gen3 = decipher_generator(gen2)
    >>>
    >>> all([a == b for a, b in zip(gen1, gen3)])
    True
    >>>
    """
    with Printer("Deciphering generator..."):
        head, gen, pack = get_header(gen=gen)
        fini = True
        if head == "aes bloc":
            yield decipher_aes(
                HEADER[head][1] + pack + b"".join(gen),
                passphrase=psw
                )
        elif head == "rsa bloc":
            yield decipher_rsa(
                HEADER[head][1] + pack + b"".join(gen),
                private_key_pem=psw
                )
        elif head == "aes sel":
            if psw == None:
                raise DecryptError("Je ne peux pas dechiffrer sans mot de passe!")
            pack = pack + b"".join(gen)
            sel, pack = pack[:32], pack[32:]
            yield decipher_aes(
                pack,
                passphrase=hashlib.blake2s(
                    sel + psw.encode("utf-8"),
                    digest_size=32
                    ).digest()
                )
        elif head == "aes gen":
            if psw == None:
                raise DecryptError("Je ne peux pas dechiffrer sans mot de passe!")
            while len(pack) < 32:
                pack += next(gen)
            sel, pack = pack[:32], pack[32:]
            hash_ = hashlib.sha512(sel + psw.encode("utf-8")).digest()
            passphrases = [hash_[:32], hash_[32:]]
            fini = False
        elif head == "rsa gen":
            size, gen, pack = tag_to_size(gen=gen, pack=pack)
            while len(pack) < size:
                pack += next(gen)
            hash_, pack = decipher_rsa(
                pack[:size],
                private_key_pem=psw
                ), \
                pack[size:]
            passphrases = [hash_[32*i:32*(i+1)] for i in range(len(hash_)//32)]
            fini = False
        else:
            raise ValueError(
                "L'entete n'est pas correcte. Elle vaut %s." % head)

        if not fini:
            for i, pack in enumerate(deconcat_gen(gen=gen, pack=pack)):
                yield decipher_aes(
                    pack,
                    passphrase=passphrases[i%len(passphrases)]
                    )

def _decipher_file(path, **kwargs):
    """
    |=======================|
    | Dechiffre un fichier. |
    |=======================|
    
    Supprime l'ancien fichier pour ne laisser
    que le nouveau fichier dechiffre.
    Ne fait aucune verifcation sur les entrees.

    parametres
    ----------
    :param filename: Le nom d'un fichier ou son chemin.
    :type filename: str
    
    Les autres arguments sont les memes que
    ceux de la fonction 'decipher_gen'.

    sortie
    ------
    :return: Le nom du fichier dechiffre.
    :rtype: str

    exemple
    -------
    :Example:
    >>> from raisin.security.decrypt import _decipher_file
    >>> _decipher_file("monfichier.txt.crais")
    'monfichier.txt'
    >>>
    """
    def generator(filename):
        """
        Cede filename par bout de BUFFER_SIZE.
        """
        with open(filename, "rb") as f:
            while 1:
                pack = f.read(BUFFER_SIZE)
                if pack == b"":
                    break
                yield pack
    
    with Printer("Decryption of the file %s..." % repr(path)):
        if filename[-6:] == ".crais":
            new_file = filename[:-6]
        else:
            new_file = "deciphered_" + filename
        new_file = filename + ".crais"
        try:
            with open(new_file, "wb") as f:
                for pack in decipher_gen(
                        generator(filename),
                        **kwargs):
                    f.write(pack)
        except Exception as e:
            if os.path.exists(new_file):
                os.remove(new_file)
            raise e from e
        return new_file

def authenticity(gen):
    """
    |=============================================|
    | S'assure que le generateur est authentique. |
    |=============================================|

    L'entete doit etre presente dans 'gen'.

    parametres
    ----------
    :param generator: Generateur qui cede des 'bytes'.
    :type generator: iter

    sortie
    ------
    :return: Les donnees initiales.
    :rtype: <type generator>
    """
    with Printer("Verify authenticity..."):
        head, gen, pack = get_header(gen=gen)
        if head != "authenticity":
            raise ValueError(
                "L'entete n'est pas correcte. Elle vaut %s. " % head \
                + "Or 'authenticity' est attendu.")
        
        hash_ = b""
        for is_end, pack in anticipate(deconcat_gen(gen=gen, pack=pack)):
            if is_end:
                if hash_ != pack:
                    raise DecryptError("Les donnees ne sont pas authentiques!")
                break
            hash_ = hashlib.sha512(hash_ + pack).digest()
            yield pack



#!/usr/bin/env python3

"""
|==============================|
| Gere la saisie des horaires. |
|==============================|

-Decore une frame pour interagir elegament avec un utilisateur.
"""

import copy
import datetime
import re
import time

from ...tools import Printer
from . import theme, checks
from . import tkinter, messagebox

class Schedules:
    """
    |========================================|
    | Permet de voir la variable 'schedules' |
    | comme une fonction continue.           |
    |========================================|

    exemple
    -------
    :Example:
    >>> from raisin.application.hmi.schedules import Schedules
    >>> s = Schedules({
        'friday': {'22:00': True, '8:00': False}, 
        'monday': {'22:00': True, '8:00': False},
        'saturday': {'10:00': False, '23:00': True},
        'sunday': {'10:00': False, '23:00': True},
        'thursday': {'22:00': True, '8:00': False},
        'tuesday': {'22:00': True, '8:00': False},
        'wednesday': {'22:00': True, '8:00': False}})
    >>>
    """
    def __init__(self, schedules):
        """
        parametres
        ----------
        :param schedules: Dictionaire qui contient les horaires et les valeurs.
        :type schedules: dict
        """
        import raisin.application.hmi.checks as checks
        assert checks.schedules_verification(schedules), \
            "'schedules' n'est pas correcte"
        
        self.schedules = schedules
        self.days = ["monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"]
        # self.next_day = {self.days[i]: self.days[(i+1)%len(self.days)] for i in range(len(self.days))}
        # self.previous_day = {v:c for c,v in self.next_day.items()}
        self.values = {self._to_min(d, *map(int, date.split(":"))): value
            for d, dic in self.schedules.items()
            for date, value in dic.items()} # Associ la valeur sur une echelle lineaire en minute.
        self.keys = sorted(self.values.keys()) # Met les clefs dans l'ordre.

    def __call__(self, date=None, *, day=None):
        """
        |============================|
        | Calcul l'interpolation     |
        | d'orde 0 au point demande. |
        |============================|

        parametres
        ----------
        :param date: Moment de l'interpolation. Peut etre:
            -None, Interpole avec l'heure actuelle.
            -float, int, The current time in seconds since the Epoch.
            -time.struct_time, Prend l'heure donee.
            -str, format %H:%M. Il faut preciser le jour.
        :type date: None
        :param day: Le jour en anglais. seulement si type(date) == str.
        :type day: str

        sortie
        ------
        :return: La valeur au point demande.
        :rtype: bool, int, float

        exemple
        -------
        :Example:
        >>> s()
        True
        >>> s(1608547486.959979)
        False
        >>> s("22:15", day="monday")
        True
        """
        with Printer("Search value for %s schedules..." % repr(date)) as p:
            # Parsing de la date.
            if date == None:
                date = time.gmtime()
            elif isinstance(date, float) or isinstance(date, int):
                date = time.gmtime(date)
            elif isinstance(date, str):
                assert isinstance(day, str), \
                    "'day' doit etre str, pas %s." % type(day).__name__
                assert day.lower() in self.days
                assert re.fullmatch(r"\d{1,2}:\d{1,2}", date)
                date = time.strptime("{}:{}".format(day.lower(), date), "%A:%H:%M")
            elif isinstance(date, time.struct_time):
                pass
            else:
                raise ValueError("'date' n'est pas correcte.")
            d, h, m = self.days[date.tm_wday], date.tm_hour, date.tm_min
            p.show("date: %s %02dh%02d" % (d, h, m))

            # Recherche de la date a gauche la plus proche
            # Il y a moyen d'optimiser pour etre en log(n) au lieu de n.
            # La variable nex, est la si on veut faire une interpolation d'odre 1.
            minute = self._to_min(d, h, m)
            rang = len(self.keys)
            for i, key in enumerate(self.keys):
                if key > minute: # Inegalite stricte pour tomber bien quand on fait rang-1.
                    rang = i
                    break
            if rang == 0 or rang == len(self.keys):
                pre, nex = len(self.keys)-1, 0
            else:
                pre, nex = rang-1, rang

            # Valeur de retour
            value = self.values[self.keys[pre]]
            p.show("value: %s" % value)
            return value

    def _to_min(self, d, h, m):
        """
        Retourne une date en minute relative au debut de la semaine.
        """
        return m + 60*h + 24*60*self.days.index(d)

def padding(frame, schedules_var, *, ylabel="value", converter=eval):
    """
    |===========================================|
    | Complete graphiquement un                 |
    | espace qui permet de choisir les horaires |
    | et les valeurs associees.                 |
    |===========================================|

    parametres
    ----------
    :param frame: Frame tkinter vierge a habiller.
    :type frame: tkinter.Frame
    :param paths_var: Variable tkinter dans la quelle on y
        lit et y ecrit les resultats. Son contenu doit prendre
        la forme suivante:
        ("{'friday': {'22:00': True, '8:00': False}, 'monday': {'22:00': True, '8:00': "
         "False}, 'saturday': {'10:00': False, '23:00': True}, 'sunday': {'10:00': "
         "False, '23:00': True}, 'thursday': {'22:00': True, '8:00': False}, "
         "'tuesday': {'22:00': True, '8:00': False}, 'wednesday': {'22:00': True, "
         "'8:00': False}}")
    :type paths_var: tkinter.StringVar
    :param ylabel: Nom de l'axe des ordonees.
    :type ylabel: str
    :param converter: Fonction de validation qui permet de forcer l'utilisateur a
        saisir une donne correcte. Cette fonction prend en entree la chaine de caractere
        saisie par l'utilisateur. Renvoi la donnee refletee par cette chaine.
        Leve une exception si les donnee d'entree sont incorrectes.
    :type converter: callable

    sortie
    ------
    :return: La frame habillee.
    :rtype: tkinter.Frame

    exemple
    -------
    :Example:
    >>> import tkinter
    >>> from raisin.application.hmi import schedules
    >>> master = tkinter.Tk()
    >>> var = tkinter.StringVar()
    >>> var.set(str({
        'friday': {'22:00': True, '8:00': False}, 
        'monday': {'22:00': True, '8:00': False},
        'saturday': {'10:00': False, '23:00': True},
        'sunday': {'10:00': False, '23:00': True},
        'thursday': {'22:00': True, '8:00': False},
        'tuesday': {'22:00': True, '8:00': False},
        'wednesday': {'22:00': True, '8:00': False}}))
    >>> schedules.padding(tkinter.Frame(master), var).pack(fill="both", expand=True)
    >>> master.mainloop()
    """
    global schedules, schedules_manager

    def graph(day):
        """
        Genere un canevas qui contient
        le graphique pour ce jour.
        """
        # Calcul des points
        dates = sorted({time.strptime("%s:%s" % (day, d), "%A:%H:%M") for d in list(schedules[day].keys()) + ["00:00", "23:59"]})
        values = [schedules_manager(d) for d in dates]
        for i in range(len(dates)-1): # Fonction escalier. Pour ne pas faire d'interpolation lineaire.
            dates.insert(2*i + 1, dates[2*i + 1])
            values.insert(2*i + 1, values[2*i])

        # Affichage
        import matplotlib
        import matplotlib.figure
        import matplotlib.backends.backend_tkagg
        fig = matplotlib.figure.Figure(facecolor=theme.JAUNE)                   # couleur de fond autour du graph
        ax = fig.add_subplot()
        ax.set_facecolor(theme.JAUNE)                                           # couleur de fond dans le graph
        ax.set_xlabel("time (hour)")
        ax.set_ylabel(ylabel)
        xfmt = matplotlib.dates.DateFormatter("%H:%M")                          # format de l'heure a afficher
        ax.xaxis.set_major_formatter(xfmt)
        ax.plot([datetime.datetime(*d[:6]) for d in dates], values, "o-", color=theme.POURPRE)

        return matplotlib.backends.backend_tkagg.FigureCanvasTkAgg(fig, master=frames[day]).get_tk_widget()

    def disp_entry(day):
        """
        Ecrit dans la barre de saisie du jour correspondant,
        le nouveau contenu associe a la variable schedules[day]
        """
        keys = sorted(schedules[day].keys(), key=lambda h: 60*int(h.split(":")[0]) + int(h.split(":")[1]))
        entrees[day].delete(0, "end")
        entrees[day].insert(0, ", ".join(
            "%s: %s" % (d.replace(":", "h"), schedules[day][d])
            for d in keys))

    def validate(event):
        """
        Tente de comprendre ce qu'il y a dans la zonne de saisie.
        Si ce qui est saisi est correcte, Les graphiques sont mis a jour.
        Et le contenu de la variable 'schedules_var' est mise a jour.
        """
        global schedules, schedules_manager

        with Printer("Check the input...") as p:
            # Verifications
            bloc = r"(\s*\d{1,2}[h:]\d{1,2}(?:(?:\s*:\s*)|\s+)[a-zA-Z0-9\.]+)"
            patern = bloc + r"(?:\s*[,;]" + bloc + r")*" # Model regex d'une saisie corecte.
            if not re.fullmatch(patern, event.widget.get()): # On tente de recuperer les groupes.
                p.show("Syntaxe incorrecte.") # Si la saisie n'est pas corecte,
                return # On ne va pas plus loin.
            p.show("Syntaxe correct.")

            # Recuperation du jour.
            for day, entry in entrees.items():
                if event.widget is entry:
                    break
            p.show("Day: %s." % day)
            canvas_entree[day].delete("all")
            canvas_entree[day].create_image(8, 8, image=theme.icons.refresh())
            
            # Extraction des informations.
            nouv_schedules = copy.copy(schedules)
            nouv_schedules[day] = {}
            for group in re.compile(bloc).findall(event.widget.get()):
                decomposition = re.search(
                    r"(?P<heures>\d+)[h:](?P<minutes>\d+).+?(?P<value>[a-zA-Z0-9\.]+)",
                    group)
                heures = int(decomposition.group("heures"))
                minutes = int(decomposition.group("minutes"))
                try:
                    value = converter(decomposition.group("value"))
                except KeyboardInterrupt as e:
                    raise e from e
                except Exception as e:
                    p.show("La valeur %s n'est pas pythonisable." \
                        % repr(decomposition.group("value")))
                    p.show(" and ".join(e.args))
                    canvas_entree[day].delete("all")
                    canvas_entree[day].create_image(8, 8, image=theme.icons.error())
                    return
                p.show("%02dh%02d: %s" % (heures, minutes, value))
                nouv_schedules[day]["%02d:%02d" % (heures, minutes)] = value

            # Deuxieme verification.
            if nouv_schedules == schedules:
                p.show("Nothing change.")
                canvas_entree[day].delete("all")
                canvas_entree[day].create_image(8, 8, image=theme.icons.ok())
                return
            if not checks.schedules_verification(nouv_schedules):
                o.show("Invalid data.")
                canvas_entree[day].delete("all")
                canvas_entree[day].create_image(8, 8, image=theme.icons.error())
                return

            # Enregistrement des changements.
            schedules = nouv_schedules
            schedules_manager = Schedules(schedules)
            schedules_var.set(str(schedules))

            # Visualisation des changements.
            disp_entry(day)
            for day_ in canvas_graph: # Un changement sur un jour,
                canvas_graph[day_].delete("all") # peut impacter le graphique des jours suivants.
                canvas_graph[day_] = graph(day_) # Donc dans le doute, on les met tous a jour.
                canvas_graph[day_].grid(row=0, column=0, columnspan=2, sticky="nesw")
            canvas_entree[day].delete("all")
            canvas_entree[day].create_image(8, 8, image=theme.icons.ok())

    # Verifications.
    assert isinstance(frame, tkinter.Frame), \
        "'frame' doit etre de type tkinter.Frame, pas %s." % type(frame).__name__
    assert isinstance(schedules_var, tkinter.StringVar), \
        "'schedules_var' doit etre de type tkinter.StringVar, pas %s." % type(schedules_var).__name__
    assert isinstance(ylabel, str), \
        "'ylabel' doit etre de type str, pas %s." % type(ylabel).__name__
    assert ylabel, "'ylabel' ne doit pas etre une chaine vide."
    assert hasattr(converter, "__call__"), "'converter' doit etre un callable."

    # Declaration des variables.
    frames = {} # Contient les frames pour chaque jour.
    try:
        schedules = eval(schedules_var.get())
    except (SyntaxError, TypeError) as e:
        raise ValueError(
            "'schedules' n'a pas un contenu qui represente une variable python.") from e
    assert checks.schedules_verification(schedules), \
        "'schedules' ne respecte pas les conventions."
    canvas_graph = {}
    canvas_entree = {}
    entrees = {}
    schedules_manager = Schedules(schedules)

    with Printer("Schedules padding..."):
        # Initialisation des widgets.
        frame.columnconfigure(0, weight=1)
        frame.rowconfigure(0, weight=1)
        theme.theme(frame)

        notebook = theme.theme(tkinter.ttk.Notebook(frame))
        notebook.grid(row=0, column=0, sticky="nesw")

        for day in ("monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"):
            frames[day] = theme.theme(tkinter.Frame(frame))
            notebook.add(frames[day], text=day)
            frames[day].rowconfigure(0, weight=31)
            frames[day].rowconfigure(1, weight=1)
            frames[day].rowconfigure(2, weight=1)
            frames[day].columnconfigure(0, weight=1)
            frames[day].columnconfigure(1, weight=32)
            canvas_graph[day] = graph(day)
            canvas_graph[day].grid(row=0, column=0, columnspan=2, sticky="nesw")
            canvas_entree[day] = theme.theme(tkinter.Canvas(frames[day]))
            canvas_entree[day].grid(row=1, column=0)

            entrees[day] = theme.theme(tkinter.Entry(frames[day]))
            entrees[day].grid(row=1, column=1, sticky="ew")
            entrees[day].bind("<KeyRelease>", validate)
            disp_entry(day)

        return frame

def get_schedules(schedules, *, existing_window=None, ylabel="value", converter=eval):
    """
    |======================================|
    | Interface pour changer les horaires. |
    |======================================|

    Utilise dans la mesure du possible, une jolie interface graphique.
    Les verifications sont faites ici.
    Ne modifie pas le contenu de 'schedules', fait une copie veritable.

    parametres
    ----------
    :param schedules: Les horaires deja existantes.
    :type schedules: dict
    :param existing_window: Une fenetre tkinter presente par ailleur.
    :param ylabel: Nom de l'axe des ordonees.
    :type ylabel: str
    :param converter: Fonction de validation qui permet de forcer l'utilisateur a
        saisir une donne correcte. Cette fonction prend en entree la chaine de caractere
        saisie par l'utilisateur. Renvoi la donnee refletee par cette chaine.
        Leve une exception si les donnee d'entree sont incorrectes.
    :type converter: callable

    sortie
    ------
    :return: Les nouvelles horaires.
    :rtype: dict

    exemple
    -------
    :Example:
    >>> from raisin.application.hmi import schedules
    >>> s = {
        'friday': {'22:00': True, '8:00': False}, 
        'monday': {'22:00': True, '8:00': False},
        'saturday': {'10:00': False, '23:00': True},
        'sunday': {'10:00': False, '23:00': True},
        'thursday': {'22:00': True, '8:00': False},
        'tuesday': {'22:00': True, '8:00': False},
        'wednesday': {'22:00': True, '8:00': False}}
    >>> schedules.get_schedules(s)
    {'friday': {'22:00': True, '8:00': False}, 
    'monday': {'22:00': True, '8:00': False},
    'saturday': {'10:00': False, '23:00': True},
    'sunday': {'10:00': False, '23:00': True},
    'thursday': {'22:00': True, '8:00': False},
    'tuesday': {'22:00': True, '8:00': False},
    'wednesday': {'22:00': True, '8:00': False}}
    >>>
    """
    class DialogWindow:
        """
        Permet d'interragir graphiquement avec l'utilisateur.
        """
        def __init__(self, schedules, ylabel, converter, existing_window):
            self.schedules = copy.copy(schedules)
            self.ylabel = ylabel
            self.converter = converter
            self.existing_window = existing_window
            self.violently_closed = True        # permet de savoir si la fenetre a ete fermee violement ou via le bouton 'valider'
            self.answer = None                  # reponse de l'utilisateur

            # preparation de la fenetre
            if self.existing_window:                                                        # si il y a deja une fenetre ouverte
                self.window = tkinter.Toplevel(self.existing_window)                        # on est oblige d'utiliser un toplevel sinon ca plante
                self.window.grab_set()                                                      # on fige la fenetre parente
                self.window.protocol("WM_DELETE_WINDOW", lambda : (self.window.destroy(), self.window.quit()))# il se trouve que ca semble fonctionner comme ca...
            else:                                                                           # dans le cas ou aucune instance tkinter n'existe
                self.window = tkinter.Tk()                                                  # et bien on en cre une tout simplement

            # configuration
            self.window.title("Change schedules")
            self.window.configure(background=theme.JAUNE)
            self.window.columnconfigure(0, weight=1)                                         # numero de colone, etirement relatif: On rend l'onglet redimenssionable sur la largeur
            self.window.rowconfigure(0, weight=30)
            self.window.rowconfigure(1, weight=1)
            self.window.focus_force()
            self.window.bind("<Return>", lambda event: self.quit())
            self.window.bind("<Escape>", lambda event: self.quit())

            # initialisation des variables
            self.schedules_var = tkinter.StringVar()
            self.schedules_var.set(str(self.schedules))

            # remplissage + ecoute
            padding(
                tkinter.Frame(self.window),
                self.schedules_var,
                ylabel=self.ylabel,
                converter=self.converter
                ).grid(row=0, column=0, sticky="ewsn")
            theme.theme(
                tkinter.Button(self.window, text="Valider", command=lambda: self.quit())
                ).grid(row=1, column=0)
            self.window.mainloop()

        def quit(self):
            self.violently_closed = False                                                   # si on ferme la fenetre depuis cette methode, c'est que la fermeture est propre
            self.answer = eval(self.schedules_var.get())                                    # on peut allos recuperer la reponse
            self.window.destroy()                                                           # selon qu'il y ai deja une fenetre en arriere plan
            if self.existing_window:                                                        # on applique ou non la methode destroy
                self.window.quit() 

    with Printer("Configure the raisin schedules..."):
        if tkinter:
            try:
                import matplotlib
            except ImportError as e:
                raise NotImplementedError("Il faut installer matplotlib.") from e
            
            # Avec fenetre graphique.
            g = DialogWindow(schedules, ylabel, converter, existing_window=existing_window)
            if g.violently_closed:
                raise KeyboardInterrupt("La fenetre a ete fermee violement!")
            return g.answer
        
        else:
            raise NotImplementedError("Il faut installer tkinter.")



#!/usr/bin/env python3

"""
Lecture inteligente de certain types de fichiers
Permet de metre directement les donnes du fichier sous forme tres facilement exploitables
"""

__all__ = ["open_extend"]

import io
import re

from . import csvreader
from . import pyreader


def open_extend(file, mode, **kwargs):
    """
    extension de la fonction open de base
    """
    if ((mode == "") and re.match(r".+\.py$", file.lower())) or (mode == "rp"):
        return pyreader.PyFileReader(file, kwargs.get("encoding", "utf-8"))
    elif ((mode == "") and re.match(r".+\.csv$", file.lower())) or (mode == "rc"):
        return csvreader.CsvFileReader(file, kwargs.get("encoding", "utf-8"), **kwargs)

    return open(file, mode, **kwargs)



#!/usr/bin/env python3


import io
import os
import re
import sys

import raisin
from ..tools import Printer, id_


def make_analyse():
    """
    cherche les informations relatives a tous les modules installes
    retourne un dictionaire qui a chaque module, associ ces caracteristiques
    """
    with Printer("Updated information on modules..."):
        return {modulename: get_infos(modulename) for modulename in get_installed_modules()}

def get_infos(modulename):
    """
    retourne un tuple contenant 3 champs:
        -validite du module:
            -True si il est present et teste
            -None si il est present mais pas teste
            -False si il n'est pas present ou qu'il n'est pas fonctionnel
        -version:
            -(STR) ou None
        -code en python
            -True si le module est entierement code en python
            -None si on en sait rien
            -False si il y a autre chose que du python
    """
    assert type(modulename) is str, "'modulename' doit etre le nom du module (str), or il s'agit d'un (%s)." % type(modulename)

    with Printer("getting infos about '%s' ..." % modulename) as p:
        try:
            module = __import__(modulename)
        except KeyboardInterrupt as e:
            raise e from e
        except ImportError:
            p.show("Failed to import.")
            return False, None, None
        except MemoryError:
            p.show("Memory error.")
            return False, None, None
        except Exception as e:
            p.show(e)
            return False, None, None
        test = test_module(module, modulename)
        all_python = test_all_python(module)
        try:
            try:
                version = module.__version__
            except AttributeError:
                version = module.VERSION
            if type(version) is list or type(version) is tuple: # si la version est du genre (2, 5, 3)
                version = ".".join(map(str, version))           # on en fait un "2.5.3"
            if test is True:
                p.show("It works properly, version: %s." % version)
            elif test is False:
                p.show("It is all broken, version: %s." % version)
            else:
                p.show("No test available for this module, version: %s." % version)
            return test, version, all_python
        except AttributeError:
            if test is True:
                p.show("It works properly.")
            elif test is False:
                p.show("It is all broken.")
            else:
                p.show("No test available for this module.")
            return test, None, all_python
        finally:
            del module

def test_module(module, modulename):
    """
    verifi que le module 'module', nome 'modulename'
    fonctionne correctement
    retourne True si le module est ok et False si il est defaillant
    Si on en sait rien, return None
    """
    with Printer("verification of the '%s' module..." % modulename) as p:
        if modulename == "abc":
            return None
        elif modulename == "aifc":
            return None
        elif modulename == "argparse":
            return None
        elif modulename == "array":
            return None
        elif modulename == "ast":
            return None
        elif modulename == "asynchat":
            return None
        elif modulename == "asyncio":
            return None
        elif modulename == "asyncore":
            return None
        elif modulename == "atexit":
            return None
        elif modulename == "audioop":
            return None
        elif modulename == "base64":
            return None
        elif modulename == "bdb":
            return None
        elif modulename == "binascii":
            return None
        elif modulename == "binhex":
            return None
        elif modulename == "bisect":
            return None
        elif modulename == "builtins":
            return None
        elif modulename == "bz2":
            return None
        elif modulename == "calendar":
            return None
        elif modulename == "cgi":
            return None
        elif modulename == "cgitb":
            return None
        elif modulename == "chunk":
            return None
        elif modulename == "cmath":
            return None
        elif modulename == "cmd":
            return None
        elif modulename == "code":
            return None
        elif modulename == "codecs":
            return None
        elif modulename == "codeop":
            return None
        elif modulename == "collections":
            return None
        elif modulename == "colorsys":
            return None
        elif modulename == "compileall":
            return None
        elif modulename == "concurrent":
            return None
        elif modulename == "configparser":
            return None
        elif modulename == "contextlib":
            return None
        elif modulename == "copy":
            return None
        elif modulename == "copyreg":
            return None
        elif modulename == "cProfile":
            return None
        elif modulename == "csv":
            return None
        elif modulename == "ctypes":
            return None
        elif modulename == "datetime":
            return None
        elif modulename == "dbm":
            return None
        elif modulename == "decimal":
            return None
        elif modulename == "difflib":
            return None
        elif modulename == "dis":
            return None
        elif modulename == "distutils":
            return None
        elif modulename == "doctest":
            return None
        elif modulename == "dummy_threading":
            return None
        elif modulename == "email":
            return None
        elif modulename == "encodings":
            return None
        elif modulename == "enum":
            return None
        elif modulename == "errno":
            return None
        elif modulename == "faulthandler":
            return None
        elif modulename == "filecmp":
            return None
        elif modulename == "fileinput":
            return None
        elif modulename == "fnmatch":
            return None
        elif modulename == "formatter":
            return None
        elif modulename == "fractions":
            return None
        elif modulename == "ftplib":
            return None
        elif modulename == "functools":
            return None
        elif modulename == "gc":
            return None
        elif modulename == "getopt":
            return None
        elif modulename == "getpass":
            return None
        elif modulename == "gettext":
            return None
        elif modulename == "glob":
            return None
        elif modulename == "gzip":
            return None
        elif modulename == "hashlib":
            return None
        elif modulename == "heapq":
            return None
        elif modulename == "hmac":
            return None
        elif modulename == "html":
            return None
        elif modulename == "http":
            return None
        elif modulename == "imaplib":
            return None
        elif modulename == "imghdr":
            return None
        elif modulename == "imp":
            return None
        elif modulename == "importlib":
            return None
        elif modulename == "inspect":
            return None
        elif modulename == "io":
            return None
        elif modulename == "ipaddress":
            return None
        elif modulename == "itertools":
            return None
        elif modulename == "json":
            return None
        elif modulename == "keyword":
            return None
        elif modulename == "lib2to3":
            return None
        elif modulename == "linecache":
            return None
        elif modulename == "locale":
            return None
        elif modulename == "logging":
            return None
        elif modulename == "lzma":
            return None
        elif modulename == "macpath":
            return None
        elif modulename == "mailbox":
            return None
        elif modulename == "mailcap":
            return None
        elif modulename == "marshal":
            return None
        elif modulename == "math":
            return None
        elif modulename == "mimetypes":
            return None
        elif modulename == "mmap":
            return None
        elif modulename == "modulefinder":
            return None
        elif modulename == "multiprocessing":
            return None
        elif modulename == "netrc":
            return None
        elif modulename == "nntplib":
            return None
        elif modulename == "numbers":
            return None
        elif modulename == "operator":
            return None
        elif modulename == "optparse":
            return None
        elif modulename == "os":
            return None
        elif modulename == "parser":
            return None
        elif modulename == "pathlib":
            return None
        elif modulename == "pdb":
            return None
        elif modulename == "pickle":
            return None
        elif modulename == "pickletools":
            return None
        elif modulename == "pkgutil":
            return None
        elif modulename == "platform":
            return None
        elif modulename == "plistlib":
            return None
        elif modulename == "poplib":
            return None
        elif modulename == "pprint":
            return None
        elif modulename == "profile":
            return None
        elif modulename == "pstats":
            return None
        elif modulename == "py_compile":
            return None
        elif modulename == "pyclbr":
            return None
        elif modulename == "pydoc":
            return None
        elif modulename == "queue":
            return None
        elif modulename == "quopri":
            return None
        elif modulename == "random":
            return None
        elif modulename == "re":
            if re.fullmatch(r"\s", " "):
                return True
            return False
        elif modulename == "reprlib":
            return None
        elif modulename == "rlcompleter":
            return None
        elif modulename == "runpy":
            return None
        elif modulename == "sched":
            return None
        elif modulename == "select":
            return None
        elif modulename == "selectors":
            return None
        elif modulename == "shelve":
            return None
        elif modulename == "shlex":
            return None
        elif modulename == "shutil":
            return None
        elif modulename == "signal":
            return None
        elif modulename == "site":
            return None
        elif modulename == "smtpd":
            return None
        elif modulename == "smtplib":
            return None
        elif modulename == "sndhdr":
            return None
        elif modulename == "socket":
            return None
        elif modulename == "socketserver":
            return None
        elif modulename == "sqlite3":
            return None
        elif modulename == "ssl":
            return None
        elif modulename == "stat":
            return None
        elif modulename == "statistics":
            return None
        elif modulename == "string":
            return None
        elif modulename == "stringprep":
            return None
        elif modulename == "struct":
            return None
        elif modulename == "subprocess":
            return None
        elif modulename == "sunau":
            return None
        elif modulename == "symbol":
            return None
        elif modulename == "symtable":
            return None
        elif modulename == "sys":
            return None
        elif modulename == "sysconfig":
            return None
        elif modulename == "tabnanny":
            return None
        elif modulename == "tarfile":
            return None
        elif modulename == "telnetlib":
            return None
        elif modulename == "tempfile":
            return None
        elif modulename == "textwrap":
            return None
        elif modulename == "threading":
            return None
        elif modulename == "time":
            return None
        elif modulename == "timeit":
            return None
        elif modulename == "tkinter":
            return None
        elif modulename == "token":
            return None
        elif modulename == "tokenize":
            return None
        elif modulename == "trace":
            return None
        elif modulename == "traceback":
            return None
        elif modulename == "tracemalloc":
            return None
        elif modulename == "turtle":
            return None
        elif modulename == "types":
            return None
        elif modulename == "typing":
            return None
        elif modulename == "unicodedata":
            return None
        elif modulename == "unittest":
            return None
        elif modulename == "urllib":
            return None
        elif modulename == "uu":
            return None
        elif modulename == "uuid":
            return None
        elif modulename == "venv":
            return None
        elif modulename == "warnings":
            return None
        elif modulename == "wave":
            return None
        elif modulename == "weakref":
            return None
        elif modulename == "webbrowser":
            return None
        elif modulename == "wsgiref":
            return None
        elif modulename == "xdrlib":
            return None
        elif modulename == "xml":
            return None
        elif modulename == "xmlrpc":
            return None
        elif modulename == "zipapp":
            return None
        elif modulename == "zipfile":
            return None
        elif modulename == "zipimport":
            return None
        elif modulename == "zlib":
            return None
        else:
            p.show("the module %s is unknown" % modulename)
            return None

def test_all_python(module):
    """
    retourne True si le module 'module' est entierement code en python
    retourne False le cas echeant
    """
    return False # a coder

def get_installed_modules():
    """
    cede le nom de tous les modules present sur cette machine
    """
    with Printer("Recovery of the names of all installed modules..."):
        for dossier in sys.path:                    # pour chaque repertoir ou python est succeptible d'aller chercher des modules
            if os.path.isdir(dossier) and dossier != "":
                for file in os.listdir(dossier):    # pour chaque machine qu'il y a dans ce repertoire
                    if os.path.isfile(os.path.join(dossier, file)) and re.fullmatch(r".+\.py", file):# si c'est un fichier python
                        if re.fullmatch(r"[a-zA-Z_\x7f-\U000fffff][a-zA-Z0-9_\x7f-\U000fffff]*", file[:-3]):
                            if file[:-3] not in ["antigravity"]:# si ce n'est pas un module a eviter
                                yield file[:-3]     # alors ce fichier est un module
                    elif os.path.isdir(os.path.join(dossier, file)) and file != "__pycache__":
                        if re.fullmatch(r"[a-zA-Z_\x7f-\U000fffff][a-zA-Z0-9_\x7f-\U000fffff]*", file):
                            if file not in ["antigravity"]:# si ce n'est pas un module a eviter
                                yield file

def get_dependencies(modulename):
    """
    retourne l'ensemble des modules directement importe a l'interieur
    de 'modulename'
    """
    assert type(modulename) is str, "'modulename' doit etre le nom du module (str), or il s'agit d'un (%s)." % type(modulename)

    with Printer("Recherche des dependance du module '%s'..." % modulename) as p:
        with Printer("Recherche de l'emplacement..."):
            emplacements = []                       # contient les chemin de tous les fichier python faisant parti du module
            for papa in sys.path:                   # on regarde de partou ou sont cence se trouver les modules
                papa = os.path.abspath(papa)
                if not os.path.isdir(papa):
                    continue
                for m in os.listdir(papa):
                    if m == modulename + ".py":     # si on a trouve le module et que c'est un fichier
                        emplacements.append(os.path.join(papa, m)) # on l'ajoute
                        p.show(os.path.join(papa, m))
                    elif m == modulename and os.path.isdir(os.path.join(papa, m)): # si le module est un repertoire
                        for base, dirs, files in os.walk(os.path.join(papa, m)): # on fouille les sous-repertoires
                            if os.path.join(os.path.dirname(__file__), "venv") in base: # sauf si il s'agit du repertoir d'environement virtuel
                                continue            # car ca permet de gagner du temps si on s'analyse soit meme
                            for file in files:
                                if file.split(".")[-1] == "py":
                                    emplacements.append(os.path.join(base, file))
                                    p.show(os.path.join(base, file))
            if not emplacements:                    # si on a rien trouve
                p.show("Module file not found...")
                return set()
        with Printer("Analysis of these files..."):
            depend = set()
            for filname in emplacements:
                with raisin.open(filname, "rp") as f:
                    try:
                        tree = f.read()             # on extrait l'arbre syntaxique de chacun des ces fichiers
                    except SyntaxError:
                        break
                    depend |= tree.get_modules()    # puis on y recherche les modules
            depend = {mod.split(".")[0] for mod in depend}
            depend = {mod for mod in depend if mod and mod != modulename}
            return depend

def get_all_dependencies(modulename):
    """
    retourne l'ensemble des modules qui interviennent de facon
    plus ou moin directe dans modulename
    retourne les dependances, et les epandance des depandances, et les ...
    annalyse les dependance (afin d'eviter une boucle infinie)
    """
    assert type(modulename) is str, "'modulename' doit etre le nom du module (str), or il s'agit d'un (%s)." % type(modulename)

    def get_all(modulename, made):
        """
        renvoi les dependances et le nouveau made
        """
        deps = get_dependencies(modulename)
        made.add(modulename)
        continuer = True
        while continuer:
            continuer = False
            for dep in deps.copy():
                if dep not in made:
                    n_dep, made = get_all(dep, made)
                    if n_dep & deps:
                        continuer = True
                    deps |= n_dep
        return deps, made

    with Printer("Recherche de toutes les dependances de '%s', meme les plus elognes..." % modulename):
        deps, made = get_all(modulename, set())
        return deps

def get_unmet_dependencies(modulename):
    """
    retourne l'ensemble des dependances non satisfaite
    par le module de nom 'modulename'
    """
    assert type(modulename) is str, "'modulename' doit etre le nom du module (str), or il s'agit d'un (%s)." % type(modulename)

    missing = set()
    with Printer("Get unmet dependances..."):
        for dep in get_dependencies(modulename):
            validity, version, is_python = get_infos(dep)
            if validity is False:
                missing.add(dep)
    return missing

def install(modulename):
    """
    install le module 'modulename'
    retourne True si le module est installe et fonctionne
    retourne None si il est installe mais qu'il n'est pas teste
    retourne False si il y a un quoique
    """
    def install_tkinter():
        """
        sur linux, tkinter s'installe un peu speciallement
        """
        with Printer("Install tkinter from package manager...") as p:
            if os.name == "posix":
                p.show("Try with 'apt'.")
                if id_().has_admin:
                    p.show("Apt has administrator.")
                    if os.system("apt install python3-tk") == 0:
                        return True
                else:
                    p.show("Please put me as administrator.")
                p.show("Try with 'packman'.")
                if os.system("packman -S python3-tk") == 0:
                    return True
            return False

    def install_giacpy():
        """
        xcas est un peu complique a installer, quelque soit la situation

        si il y a le bug:
            W: Erreur de GPG : http://www-fourier.ujf-grenoble.fr/~parisse/debian stable Release : 
            Les signatures suivantes n'ont pas pu être vérifiées car la clé publique n'est pas disponible : 
            NO_PUBKEY 8E62E25D998B1156
        on peu faire:
            gpg --keyserver hkp://keyserver.ubuntu.com:11371 --recv-key 998B1156
            gpg -a --export 998B1156 | apt-key add -
            apt-get update
        """
        with Printer("Install tkinter from package manager...") as p:
            if os.name == "posix":
                p.show("Try with 'apt'.")
                if id_().has_admin:
                    p.show("Apt has administrator.")
                    if os.system("apt install python3-giacpy") == 0:
                        return True
                else:
                    p.show("Please put me as administrator.")
                p.show("Try with 'packman'.")
                if os.system("pacman -S python3-giacpy") == 0:
                    return True
                p.show("Try manualy.")
                if id_().has_admin:
                    p.show("As administrator.")
                    if os.system("add-apt-repository 'deb http://www-fourier.ujf-grenoble.fr/~parisse/debian/ stable main'") == 0:
                        with open("xcas_public_key.gpg", "wb") as f:
                            f.write(b'-----BEGIN PGP PUBLIC KEY BLOCK-----\n'
                                    b'Version: GnuPG v1.4.10 (GNU/Linux)\n\n'
                                    b'mQENBE/a6hoBCAC/ReFI2FbsyL5csp7guN5m+x'
                                    b'l4MwRdLWt00MtgkdL6oMEN+hug\nz/nmZaPoZq'
                                    b'uDIDNDPssNGuwU0nUivYGtIaT60AJ7zv7yWGhF'
                                    b'KLPcrv7Jo97crYOK\nskEE03L2zOA1DRxlZo6h'
                                    b'vUdfZj8XgjxSvyhmrMiyoP5iQf5vPhUbUGwYZ1'
                                    b'xNQpZh\nQjzFEZpuZ/QC0KkrGjSP42vAK0rEEM'
                                    b'AJ3eaHxlUxdykg6+De274AS4xv2/otB6JF\ng0'
                                    b'TWpfyI+lAo+ydhFDFecceIrZI6xNhKRXhuamfY'
                                    b'tegUVq06EGs0OZ+7S+CBSpVH\nX/TJea3HpQk2'
                                    b'ugfiKkxhZ+GYzWEWfDz8g36JABEBAAG0MUJlcm'
                                    b'5hcmQgUGFyaXNz\nZSA8YmVybmFyZC5wYXJpc3'
                                    b'NlQHVqZi1ncmVub2JsZS5mcj6JATgEEwECACIF'
                                    b'Ak/a\n6hoCGwMGCwkIBwMCBhUIAgkKCwQWAgMB'
                                    b'Ah4BAheAAAoJEI5i4l2ZixFWEwkH/0+4\nVUrU'
                                    b'GpLUEnaNa6d2ft/ccqn+JeL3/jOAjTCiirTSrM'
                                    b'hIjaVoOPt4Le69TA2Pxg9t\nWC3vI2TKSUfON1'
                                    b'YOzsVO4PlQGrQRIUPFnlouIeem4AzD6Jk0KZbW'
                                    b'qNJEnscT4Gvy\nhFg4XaXVFhv4ilNNY7Iw2qqW'
                                    b'1bzp3GDF4jUPRVJpkivsrwMSYgmOTcDpHuwB7B'
                                    b'Uh\n0+peJD0nakVBRSR86zGKKpfvjYvMSWXAXn'
                                    b'qf3Hr/2PajgWWcj78IwT6ouQWiE4a9\nMYwXfR'
                                    b'MLjJEePtvYXzs52Qvqk2avQXlac6Cj57r3XrUD'
                                    b'RTTQKLMvnUAFDOuk0K0W\n3VxL0giUwV+2E/Pp'
                                    b'yeu5AQ0ET9rqGgEIAMKv0pLzuhZAINO+QReBvM'
                                    b'wu5oByptfs\n5FOqD6NXovsml6jwlZENKvrHj2'
                                    b'+kB9cbPEzxYQe7BvVnTcg1LuD6ALfzr7MXleIH'
                                    b'\nrlZlQDIYxDy/C4ST7yk5LWz9itpOExYVMqn2'
                                    b'CmWwIK24MUrTgmo4uT25lLqDxll7\naS7w3QV3'
                                    b'YMhqGzw4Pa2lNtnbBqiFwqp8zgaVH1l5iSARpi'
                                    b'dOTlBWuWrfbdKpyuZb\nmHZOyi2qy5pSvWO333'
                                    b'Q9tQsCe1CFilc3c3rHNwO7NkU+2PE5vlTcFVj1'
                                    b't1D1tllm\nl1L3C62uqbfzkO47bbSZzp38yZPt'
                                    b'3yqQY9R2WFJnLR2mJcGR6UToY0MAEQEAAYkB\n'
                                    b'HwQYAQIACQUCT9rqGgIbDAAKCRCOYuJdmYsRVo'
                                    b'i2B/4neWZ8F6QE5Hu06eEitPbV\nH5YtZl7IyA'
                                    b'pe/X6/qROdt9tfSdMJrN4vl31LMK6ECCnGN4Vd'
                                    b'ZW13Z2D2mQqsjwDy\nM0b4lkmQ5ZHwzQLacvo1'
                                    b'XcSCazi9Av9rGQpnwu4k+vioMk1znsyVGqJgjS'
                                    b'EQDRI+\nZ9YANtGurhHYUgxT9WCJod3ogvDnrq'
                                    b'TFSRpEyrZPPDZb9HfFTI+rTHGTRHCx4+xk\nwe'
                                    b'CxL8fVj3/YDrii9FW/Hm1fSt1zwQZWgX2cHte+'
                                    b'6pD6gVH4c6h0C87yZOJ4YxcH\ntbU0JgFGJhf6'
                                    b'GxHjCk5/F5bLMHjfsVLrJyWd7sN3gH76MwcSLf'
                                    b'tF/Durjx0F9r64\n=1H7u\n-----END PGP PU'
                                    b'BLIC KEY BLOCK-----\n')
                        os.system("apt add xcas_public_key.gpg")
                        os.remove("xcas_public_key.gpg")
                        if os.system("apt update") == 0:
                            if os.system("apt-get install *giac*") == 0:
                                return True
                else:
                    p.show("Please put me as administrator.")
            return False

    assert type(modulename) is str, "'modulename' doit etre le nom du module (str), or il s'agit d'un (%s)." % type(modulename)

    with Printer("Install '%s'..." % modulename) as p:
        validity, version, is_python = get_infos(modulename)
        if validity is True: # si le module est present et teste
            p.show("The module '%s' is already installed!" % modulename)
            return validity
        elif validity is None: # si le module est present mais pas teste
            p.show("The module '%s' is already installed, but it is not tested!" % modulename)
            return validity

        if modulename == "tkinter":
            return install_tkinter()
        elif modulename == "Crypto":
            return install("pycryptodome")
        elif modulename == "Cryptodome":
            return install("pycryptodomex")
        elif modulename == "cv2":
            return install("opencv-python")
        elif modulename == "sksound":
            return install("scikit-sound")
        elif modulename == "giacpy":
            return install_giacpy()
        elif modulname == "graph":
            return install("graph-theory")

        # install with pip
        with Printer("Install '%s' with 'pip'..." % modulename) as p:
            if id_().has_admin:
                p.show("Installation with administrator privileges.")
            else:
                p.show("Installation without admin privilages.")
            res = os.system("%s -m pip -q install --upgrade %s" % (sys.executable, modulename))
            if res == 0:
                validity, version, is_python = get_infos(modulename)
                if validity in (True, None):
                    p.show("Le module '%s' a ete installe avec succes!" % modulename)
                    return validity

        p.show("Le module '%s' ne s'est pas bien installe!" % modulename)
        return False




#!/usr/bin/env python3

"""
C'est ici qu'il y a les petites fonctions
utiles un peu de partout.
Bref, il y a tout ce qui est partage.
"""

import os
import uuid

from .constants import BUFFER_SIZE, HEADER
from ..tools import temprep


def size_to_tag(size):
    r"""
    |=================================================|
    | Cre un fanion qui permet de regrouper n octets. |
    |=================================================|
    
    parametres
    ----------
    :param size: Le nombre d'octets a englober.
    :type size: int

    sortie
    ------
    :return: Les donnees chiffrees.
    :rtype: bytes

    exemple
    -------
    :Example:
    >>> from raisin.serialization.tools import size_to_tag
    >>>
    >>> data = b'\x00\xb0\xe9\xe3\xd4\xcdv\xaf\xbf\xde'
    >>> size_to_tag(len(data)) + data
    b'\x8a\x00\xb0\xe9\xe3\xd4\xcdv\xaf\xbf\xde'
    >>>
    """    
    # on va decouper ce nbr en base 2**7 pour garder un bit de libre par octet
    binaire_txt = bin(size)[2:]
    binaire_txt = "0"*(-len(binaire_txt) % 7) + binaire_txt # on bourre de 0 afin d'avoir un multiple de 7
    fanion_list = []    # contient les bytes correspondants
    for i in range(len(binaire_txt)//7): # pour chaque sous_packets
        fanion_list.append(sum((int(bit) << j for j, bit in enumerate(binaire_txt[7*i:7*i+7][::-1])))) # on ecrit le nombre en base 7
    fanion_list[-1] += 1 << 7 # on ajoute un drapeau pour dire qu'il n'y a plus rien apres
    return bytes(fanion_list) # on transforme ca en type bytes

def tag_to_size(*, gen=(lambda:(yield from []))(), pack=b""):
    r"""
    |=============================================|
    | Regroupe et retourne un packet d'octets     |
    | prealablement 'fanionne' par 'size_to_tag'. |
    |=============================================|
    Leve une StopIteration exception si pack et le generateur sont epuises.
    Leve une ValueError si les entrees sont incorrectes.
    
    parametres
    ----------
    :param gen: La suite des paquets, generateur de 'bytes'.
    :type gen: iter
    :param pack: Le debut d'une suite d'octet, fanion inclus.
    :type pack: bytes

    sortie
    ------
    :return: (
        La longeur representee par le fanion,
        Le generateur possiblement un peu plus itere,
        'pack' depourvu du fanion.
        )
    :rtype: tuple (int, iter, bytes)

    exemple
    -------
    :Example:
    >>> from raisin.serialization.tools import *
    >>>
    >>> data1 = b'\x00\xb0\xe9\xe3\xd4\xcdv\xaf\xbf\xde'
    >>> data2 = b'\xe9\xd1\xb7\x0e\xe6-N4n\xb4'
    >>> ens = size_to_tag(len(data1)) + data1 + data2
    >>> ens
    b'\x8a\x00\xb0\xe9\xe3\xd4\xcdv\xaf\xbf\xde\xe9\xd1\xb7\x0e\xe6-N4n\xb4'
    >>> tag_to_size(ens, (lambda:(yield b''))())
    (10, <generator object <lambda> at 0x7ff437890350>,
    b'\x00\xb0\xe9\xe3\xd4\xcdv\xaf\xbf\xde\xe9\xd1\xb7\x0e\xe6-N4n\xb4')
    >>>
    """
    gen = iter(gen)
    size = 0        # initialisation
    while 1:        # tant que le fanion n'est pas termine
        if pack:
            octet = pack[0]
            pack = pack[1:]
        else:
            while not pack:
                try:
                    pack = next(gen)
                except (StopIteration, TypeError) as e:
                    if size:
                        raise ValueError("L'indicateur de longueur de packet n'est pas complet.") from e
                    raise StopIteration from e
            octet, pack = pack[0], pack[1:]
        if octet > 127: # si on arrive a la fin du fanion
            size = (size << 7) + (octet - 128)
            return size, gen, pack
        size = (size << 7) + octet

def to_gen(*, gen=(lambda:(yield from []))(), pack=b""):
    r"""
    |=============================================|
    | Regroupe les packets cedes par 'gen' afin   |
    | de cedez des packets de longeur normalizee. |
    |=============================================|
    
    parametres
    ----------
    :param gen: La suite des paquets, generateur de 'bytes'.
    :type gen: iter
    :param pack: Le debut d'une suite d'octet, fanion inclus.
    :type pack: bytes

    sortie
    ------
    :return: Les packets de taille BUFFER_SIZE.
    :rtype: bytes

    exemple
    -------
    :Example:
    >>> import random
    >>> from raisin.serialization.constants import BUFFER_SIZE
    >>> from raisin.serialization.tools import gen_to_gen
    >>>
    >>> l = [b'\x00'*random.randint(0, 2*BUFFER_SIZE) for i in range(5)]
    >>> [len(e) for e in l]
    [6051175, 14551972, 13463655, 7948728, 12289630]
    >>>
    >>> lbis = list(gen_to_gen(iter(l)))
    >>> [len(e) for e in lbis]
    [10485760, 10485760, 10485760, 10485760, 10485760, 1876360]
    >>>
    >>> BUFFER_SIZE
    10485760
    >>>
    """
    for data in gen:
        pack += data
        while len(pack) > BUFFER_SIZE:
            yield pack[:BUFFER_SIZE]
            pack = pack[BUFFER_SIZE:]
    yield pack

def anticipate(frozgen, **kwargs):
    r"""
    |====================================================|
    | Permet de savoir si le packet cede est le dernier. |
    |====================================================|

    parametres
    ----------
    :param frozgen:
        -Generateur de 'longeur' inconnue,
        qui cede des objets de longeur quelquonque.
    :type frozgen: iter
    :param frozpack: Le premier packet de 'frozgen'.
        Ce n'est pas une copie de ce que va commencer par cede 'frozgen'.
        ```
            anticipate(g, frozpack=frozpack)
        <=> anticipate(iter([frozpack] + list(g)))
        ```
    :type frozpack: object

    sortie
    ------
    :return: A chaque iterations cede 2 choses:
        -Un booleen qui vaut True si le packet est le dernier, False sinon.
        -Le packet.
        Si le generateur ced un None, leve un ValueError.
    :rtype: tuple (bool, object)

    exemple
    -------
    :Example:
    >>> from raisin.serialization.tools import anticipate
    >>>
    >>> gen = [0, "a", 1.5, b'\xff']
    >>> for is_end, e in anticipate(gen):
    ...     print(is_end, e)
    ...
    False 0
    False a
    False 1.5
    True b'\xff'
    >>>
    """
    is_empty = True
    if "frozpack" in kwargs:
        prec = kwargs["frozpack"]
        is_empty = False

    for i, pack in enumerate(frozgen):
        if i == 0 and is_empty:     # si on a pas encore constitue de memoire
            prec = pack             # et bien on fait une boucle a vide
            is_empty = False        # de sorte a s'assurer qu'il y a une memoire
            continue
        yield False, prec           # si par contre il y a une memoire
        prec = pack                 # on cede puis on decale d'un cran

    if not is_empty:
        yield True, pack

def concat_gen(frozgen, *, frozpack=None):
    r"""
    |======================================|
    | Permet de concatener les elements du |
    | generateur de facon reversible.      |
    |======================================|
    
    parametres
    ----------
    :param frozgen: Generateur de 'bytes' inconnue.
    :type frozgen: iter
    :param frozpack: Le premier packet de 'frozgen'.
        Ce n'est pas une copie de ce que va commencer par cede 'frozgen'.
        ```
            concat_gen(g, frozpack=frozpack)
        <=> concat_gen(iter([frozpack] + list(g)))
        ```
    :type frozpack: bytes

    sortie
    ------
    :return: Les elements du generateurs pourvu d'un fanion.
    :rtype: bytes

    exemple
    -------
    :Example:
    >>> from raisin.serialization.tools import concat_gen
    >>>
    >>> gen = (b'\x00'*i for i in range(5))
    >>> b''.join(concat_gen(gen))
    b'\x80\x81\x00\x82\x00\x00\x83\x00\x00\x00\x84\x00\x00\x00\x00'
    >>>
    """
    if frozpack != None:
        yield size_to_tag(len(frozpack)) + frozpack
    for e in frozgen:
        yield size_to_tag(len(e)) + e

def deconcat_gen(*, gen=(lambda:(yield from []))(), pack=b""):
    r"""
    |================================|
    | Retrouve les packets initiaux. |
    |================================|
    Leve une ValueError si l'entree n'est pas correcte.
    
    parametres
    ----------
    :param gen:
        La suite des paquets, generateur de 'bytes'.
        Les packets cedes ne doivent pas specialement etre bien regroupes.
        Par contre ils doivents etre dans le bon ordre.
    :type gen: iter
    :param pack:
        pack: Le premier packet de 'gen'.
        Ce n'est pas une copie de ce que va commencer par cede 'gen'.
        ```
            deconcat_gen(g, pack=pack)
        <=> deconcat_gen(iter([pack] + list(g)))
        ```
    :type pack: bytes

    sortie
    ------
    :return: Les packets initialement encapsules..
    :rtype: bytes

    exemple
    -------
    :Example:
    >>> from raisin.serialization.tools import *
    >>> gen = iter([b'a', b'houla', b'', b'hihi'])
    >>> d = b''.join(concat_gen(gen))
    >>> d
    b'\x81a\x85houla\x80\x84hihi'
    >>> list(deconcat_gen(pack=d))
    [b'a', b'houla', b'', b'hihi']
    >>>
    """
    n_packet = 0
    while 1:
        # recuperation de la longueur
        n_packet += 1
        try:
            size, gen, pack = tag_to_size(gen=gen, pack=pack)
        except ValueError as e:
            raise ValueError("Probleme dans l'entete du packet n°%d." % n_packet) from e
        except StopIteration:
            break

        # extraction du bon nombre de donnees
        while len(pack) < size:
            try:
                pack += next(gen)
            except StopIteration as e:
                raise ValueError(
                    "Le packet n°%d est incomplet.\n" \
                    + "Longeur attendue: %d\n" % size \
                    + "Longeur disponible: %s." % len(pack)) from e

        # on cede les donnees
        yield pack[:size]
        pack = pack[size:]

def get_header(*, gen=(lambda:(yield from []))(), pack=b""):
    """
    |=================================|
    | Extrait et interprete l'entete. |
    |=================================|
    Leve une ValueError si l'entete n'est pas correcte.

    parametres
    ----------
    :param gen: Generateur de 'bytes'.
    :type gen: iter
    :param pack: Le premier packet de 'gen'.
        Ce n'est pas une copie de ce que va commencer par cede 'frozgen'.
        ```
            get_header(g, pack=pack)
        <=> get_header(iter([pack] + list(g)))
        ```
    :type pack: bytes

    sortie
    ------
    :return: (
        La signification de l'entete,
        Le generateur possiblement un peu plus itere,
        'pack' depourvu de l'entete.
        )
    :rtype: tuple (str, iter, bytes)

    exemple
    -------
    :Example:
    >>> from raisin.serialization.tools import get_header
    >>>
    >>> data = b'</>small int</>3246510'
    >>> h, g, p = get_header(pack=data)
    >>> h
    'small int'
    >>> p
    b'3246510'
    >>>
    """
    gen = iter(gen)
    size_max = 0 # taille de la plus grosse balise

    # recuperation du premier octet
    while not len(pack):
        try:
            pack += next(gen)
        except StopIteration as e:
            raise ValueError("Impossible de recuperer l'entete d'une chaine vide.") from e
    head = pack[:1]
    for sign, heads in HEADER.items():
        if head == heads[1]:
            return sign, gen, pack[1:]
        size_max = max(size_max, len(heads[0]))

    # recuperation de l'entete totale
    i_end = 6
    while len(pack) < 6 or pack[i_end-3:i_end] != b"</>":
        if i_end < len(pack):
            i_end += 1
            continue
        if len(pack) > size_max:
            raise ValueError("L'entete est trop longue.\n" \
                             + "Elle ne doit pas exceder %d octets.\n" % size_max \
                             + "Le debut de l'entete vaut %s." % pack[:2*size_max])
        try:
            pack += next(gen)
        except StopIteration as e:
            raise ValueError("L'entete n'est pas de la forme b'</>...</>'.\n" \
                             + "Le debut de l'entete vaut %s." % pack[:2*size_max])
    head = pack[:i_end]
    for sign, heads in HEADER.items():
        if head == heads[0]:
            return sign, gen, pack[i_end:]
    raise ValueError("L'entete %s n'est pas connue." % head)

def relocate(frozgen):
    """
    |=========================================================|
    | Epuise un generateur en le transferant dans un fichier. |
    | Cede la meme chose que l'entree mais decharge l'entree  |
    | avant de ceder le premier packet.                       |
    |=========================================================|
    
    parametres
    ----------
    :param frozgen: Generateur de 'bytes' inconnue.
    :type frozgen: iter

    sortie
    ------
    :return: La meme chose que 'frozengen'.
    :rtype: bytes
    """

    total_size = 0  # nbr d'octets en memoire
    tampon = []     # decharge frozgen tant que c'est pas trop gros
    sizes = []      # la taille des elements presents dans le fichier
    for e in frozgen:
        total_size += len(e)
        if total_size <= BUFFER_SIZE: # cas ou tout tient dans la ram
            tampon.append(e)
        else:       # cas ou il faut decharger la ram dans le disque
            if not sizes: # cas ou c'est la goutte qui fait deborder le vase
                f = open(os.path.join(temprep(), uuid.uuid4().hex), "wb")
            f.write(e)
            sizes.append(len(e))
    yield from tampon
    if sizes:
        f.close()
        with open(f.name, "rb") as fr:
            for size in sizes:
                yield fr.read(size)
        os.remove(f.name)



#!/usr/bin/env python3

"""
Met en forme les objets de facon a les rendre transportable,
de facon securisee et optimise.
"""

__all__ = [ "serialize", "encrypt", "compress",
            "deserialize", "decrypt", "decompress",
            "tools", "constants"]



#!/usr/bin/env python3

"""
|============================|
| Desinstalle l'application. |
|============================|

Ce fichier est a la fois un module est un script.

1) Retire raisin des applications au demarrage.
2) Supprime les raccourcis qui pointent vers raisin.
3) Supprime les repertoires crees par raisin
"""

import os
import shutil

import raisin.tools as tools # on ne fait pas d'import relatif
import raisin.application.settings as settings # de facon a ce qu'on puisse l'excecuter dirrectement


def uninstall_startup(home):
    """
    Supprime dans les fichiers de configuration,
    toutes les commandes qui excecutent une
    instance de raisin au demarrage.
    """
    assert isinstance(home, str), \
        "'home' have to be 'str', not %s." % type(home).__name__
    assert os.path.isdir(home), \
        "'home' have to be a repository. " \
        + "%s is not an existing repository." % repr(home)

    with tools.Printer(
            "Adding raisin to apps at startup for {}...".format(
            repr(os.path.basename(home)))
            ) as p:
        path_ubuntu = os.path.join(home, ".config/autostart/raisin.desktop")
        path_raspberry = os.path.join(home, ".config/lxsession/LXDE-pi/autostart")
        path_windows = os.path.join(home,
            "AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs\\Startup\\raisin.pyw")
        
        if os.path.exists(path_ubuntu):
            os.remove(path_ubuntu)
            p.show("File removed (%s)." % repr(path_ubuntu))
        elif os.path.exists(path_raspberry):
            with open(path_raspberry, "r", encoding="utf-8") as f:
                content = [l for l in f if "raisin" not in l]
            with open(path_raspberry, "w", encoding="utf-8") as f:
                f.write("".join(content))
            p.show("File modified (%s)." % repr(path_raspberry))
        elif os.path.exists(path_windows):
            os.remove(path_windows)
            p.show("File removed (%s)." % repr(path_windows))
        else:
            p.show("File not found!")

def uninstall_shortcut(home):
    """
    Supprime les alias qui pointent vers une fonctionnalite de raisin
    """
    assert isinstance(home, str), \
        "'home' have to be 'str', not %s." % type(home).__name__
    assert os.path.isdir(home), \
        "'home' have to be a repository. " \
        + "%s is not an existing repository." % repr(home)

    with tools.Printer(
            "Uninstall shortcut for {}...".format(
            repr(os.path.basename(home)))
            ) as p:
        # linux
        bashrc = os.path.join(home, ".bashrc")
        if os.path.exists(bashrc):
            p.show("Removing raisin from the '.bashrc' file")
            with open(bashrc, "r", encoding="utf-8") as f:
                lines = [l for l in f if "raisin" not in l]
                while lines and lines[-1] == "\n":
                    del lines[-1]
            with open(bashrc, "w", encoding="utf-8") as f:
                f.write("".join(lines))

        # windaube
        profile = os.path.join(home, "Documents", "profile.ps1")
        if os.path.exists(profile):
            p.show("Removing raisin from the 'profile.ps1' file")
            with open(profile, "r", encoding="utf-8") as f:
                lines = [l for l in f in "raisin" not in l]
                while lines and lines[-1] == "\n":
                    del lines[-1]
            if lines:
                with open(profile, "w", encoding="utf-8") as f:
                    f.write("".join(lines))
            else:
                os.remove(profile)

def uninstall_settings(home):
    """
    Supprime le repertoire '.raisin'.
    Supprime aussi le repertoire d'enregistrement des resultats
    si il existe.
    """
    assert isinstance(home, str), \
        "'home' have to be 'str', not %s." % type(home).__name__
    assert os.path.isdir(home), \
        "'home' have to be a repository. " \
        + "%s is not an existing repository." % repr(home)

    with tools.Printer(
            "Uninstall settings for {}...".format(
            repr(os.path.basename(home)))
            ) as p:
        recording_directory = settings.Settings(home=home)["cluster_work"]["recording_directory"]
        if os.path.exists(recording_directory):
            p.show("Deletion of %s." % repr(recording_directory))
            shutil.rmtree(recording_directory)
        raisin_path = os.path.join(home, ".raisin")
        if os.path.exists(raisin_path):
            p.show("Deletion of %s." % repr(raisin_path))
            shutil.rmtree(raisin_path)

def _list_home():
    """
    Cede les repertoires personel de
    tous les utilisateurs qui doivent
    beneficier de la desinstallation de raisin.
    """
    if tools.id_().has_admin:
        if dialog.question_binaire(
                "Voulez vous desinstaller 'raisin' "
                "pour tous les utilisateurs?",
                default=False):
            racine = "C:\\Users" if os.name == "nt" else "/home"
            for user in os.listdir(racine):
                yield os.path.join(racine, user)
            raise StopIteration
    yield os.path.expanduser("~")

def main():
    """
    Desinstalle entierement l'application raisin.
    """
    with tools.Printer("Uninstall raisin..."):
        for home in _list_home():
            uninstall_shortcut(home)
            uninstall_startup(home)
            uninstall_settings(home)
        return 0

if __name__ == "__main__":
    main()
